from datetime import timedelta

from typing import Iterable, List

from common.exceptions import LogicError
from indy_common.authorize.auth_constraints import AbstractConstraintSerializer
from indy_node.server.node_bootstrap import NodeBootstrap
from indy_node.server.txn_version_controller import TxnVersionController

from indy_node.server.validator_info_tool import ValidatorNodeInfoTool

from plenum.common.constants import VERSION, \
    ENC, RAW, CURRENT_PROTOCOL_VERSION, FORCE, ATTRIB_LABEL, IDR_CACHE_LABEL
from plenum.common.txn_util import get_type, get_payload_data, TxnUtilConfig
from plenum.common.types import f, \
    OPERATION
from plenum.common.util import get_utc_datetime
from plenum.server.node import Node as PlenumNode
from state.pruning_state import PruningState
from indy_common.config_util import getConfig
from indy_common.constants import TXN_TYPE, ATTRIB, DATA, ACTION, \
    NODE_UPGRADE, COMPLETE, FAIL, POOL_UPGRADE, POOL_CONFIG, \
    IN_PROGRESS, AUTH_RULE, AUTH_RULES
from indy_common.types import Request, SafeRequest
from indy_common.config_helper import NodeConfigHelper
from indy_node.server.client_authn import LedgerBasedAuthNr
from indy_node.server.node_authn import NodeAuthNr
from stp_core.common.log import getlogger

logger = getlogger()


class Node(PlenumNode):
    keygenScript = "init_indy_keys"
    client_request_class = SafeRequest
    TxnUtilConfig.client_request_class = Request
    _info_tool_class = ValidatorNodeInfoTool

    def __init__(self,
                 name,
                 clientAuthNr=None,
                 ha=None,
                 cliname=None,
                 cliha=None,
                 config_helper=None,
                 ledger_dir: str = None,
                 keys_dir: str = None,
                 genesis_dir: str = None,
                 plugins_dir: str = None,
                 node_info_dir: str = None,
                 primaryDecider=None,
                 pluginPaths: Iterable[str] = None,
                 storage=None,
                 config=None,
                 bootstrap_cls=NodeBootstrap):
        config = config or getConfig()

        config_helper = config_helper or NodeConfigHelper(name, config)

        ledger_dir = ledger_dir or config_helper.ledger_dir
        keys_dir = keys_dir or config_helper.keys_dir
        genesis_dir = genesis_dir or config_helper.genesis_dir
        plugins_dir = plugins_dir or config_helper.plugins_dir
        node_info_dir = node_info_dir or config_helper.node_info_dir

        # TODO: 4 ugly lines ahead, don't know how to avoid
        # self.idrCache = None
        # self.attributeStore = None
        self.upgrader = None
        self.restarter = None
        self.poolCfg = None

        super().__init__(name=name,
                         clientAuthNr=clientAuthNr,
                         ha=ha,
                         cliname=cliname,
                         cliha=cliha,
                         config_helper=config_helper,
                         ledger_dir=ledger_dir,
                         keys_dir=keys_dir,
                         genesis_dir=genesis_dir,
                         plugins_dir=plugins_dir,
                         node_info_dir=node_info_dir,
                         pluginPaths=pluginPaths,
                         storage=storage,
                         config=config,
                         bootstrap_cls=bootstrap_cls)

        # TODO: ugly line ahead, don't know how to avoid
        self.clientAuthNr = clientAuthNr or self.defaultAuthNr()

        self.nodeMsgRouter.routes[Request] = self.processNodeRequest
        self.nodeAuthNr = self.defaultNodeAuthNr()
        self.db_manager.set_txn_version_controller(TxnVersionController())

    @property
    def attributeStore(self):
        return self.db_manager.get_store(ATTRIB_LABEL)

    @property
    def idrCache(self):
        return self.db_manager.get_store(IDR_CACHE_LABEL)

    def on_inconsistent_3pc_state(self):
        timeout = self.config.INCONSISTENCY_WATCHER_NETWORK_TIMEOUT
        logger.warning("Suspecting inconsistent 3PC state, going to restart in {} seconds".format(timeout))

        now = get_utc_datetime()
        when = now + timedelta(seconds=timeout)
        self.restarter.requestRestart(when)

    def getIdrCache(self):
        return self.idrCache

    def post_txn_from_catchup_added_to_domain_ledger(self, txn):
        pass

    def postPoolLedgerCaughtUp(self, **kwargs):
        # The only reason to override this is to set the correct node id in
        # the upgrader since when the upgrader is initialized, node might not
        # have its id since it maybe missing the complete pool ledger.
        self.upgrader.nodeId = self.id
        super().postPoolLedgerCaughtUp(**kwargs)

    def postConfigLedgerCaughtUp(self, **kwargs):
        self.poolCfg.processLedger()
        self.upgrader.processLedger()
        super().postConfigLedgerCaughtUp(**kwargs)
        self.acknowledge_upgrade()

    def preLedgerCatchUp(self, ledger_id):
        super().preLedgerCatchUp(ledger_id)

        if len(self.idrCache.un_committed) > 0:
            raise LogicError('{} idr cache has uncommitted txns before catching up ledger {}'.format(self, ledger_id))

    def postLedgerCatchUp(self, ledger_id):
        if len(self.idrCache.un_committed) > 0:
            raise LogicError('{} idr cache has uncommitted txns after catching up ledger {}'.format(self, ledger_id))

        super().postLedgerCatchUp(ledger_id)

    def acknowledge_upgrade(self):
        if not self.upgrader.should_notify_about_upgrade_result():
            return
        lastUpgradeVersion = self.upgrader.lastActionEventInfo.data.version
        action = COMPLETE if self.upgrader.didLastExecutedUpgradeSucceeded else FAIL
        logger.info('{} found the first run after upgrade, sending NODE_UPGRADE {} to version {}'.format(
            self, action, lastUpgradeVersion))
        op = {
            TXN_TYPE: NODE_UPGRADE,
            DATA: {
                ACTION: action,
                VERSION: lastUpgradeVersion.full
            }
        }
        op[f.SIG.nm] = self.wallet.signMsg(op[DATA])

        request = self.wallet.signRequest(
            Request(operation=op, protocolVersion=CURRENT_PROTOCOL_VERSION))

        self.startedProcessingReq(request.key, self.nodestack.name)
        self.send(request)
        self.upgrader.notified_about_action_result()

    def notify_upgrade_start(self):
        scheduled_upgrade_version = self.upgrader.scheduledAction.version
        action = IN_PROGRESS
        logger.info('{} is about to be upgraded, '
                    'sending NODE_UPGRADE {} to version {}'.format(self, action, scheduled_upgrade_version))
        op = {
            TXN_TYPE: NODE_UPGRADE,
            DATA: {
                ACTION: action,
                VERSION: scheduled_upgrade_version.full
            }
        }
        op[f.SIG.nm] = self.wallet.signMsg(op[DATA])

        # do not send protocol version before all Nodes support it after Upgrade
        request = self.wallet.signRequest(
            Request(operation=op, protocolVersion=CURRENT_PROTOCOL_VERSION))

        self.startedProcessingReq(request.key,
                                  self.nodestack.name)
        self.send(request)

    def processNodeRequest(self, request: Request, frm: str):
        if request.operation[TXN_TYPE] == NODE_UPGRADE:
            try:
                self.nodeAuthNr.authenticate(request.operation[DATA],
                                             request.identifier,
                                             request.operation[f.SIG.nm])
            except BaseException as ex:
                logger.warning('The request {} failed to authenticate {}'
                               .format(request, repr(ex)))
                return
        if not self.isProcessingReq(request.key):
            self.startedProcessingReq(request.key, frm)
        # If not already got the propagate request(PROPAGATE) for the
        # corresponding client request(REQUEST)
        self.recordAndPropagate(request, frm)

    def validateNodeMsg(self, wrappedMsg):
        msg, frm = wrappedMsg
        if all(attr in msg.keys()
               for attr in [OPERATION, f.IDENTIFIER.nm, f.REQ_ID.nm]) \
                and msg.get(OPERATION, {}).get(TXN_TYPE) == NODE_UPGRADE:
            cls = self.client_request_class
            cMsg = cls(**msg)
            return cMsg, frm
        else:
            return super().validateNodeMsg(wrappedMsg)

    def authNr(self, req):
        # TODO: Assumption that NODE_UPGRADE can be sent by nodes only
        if req.get(OPERATION, {}).get(TXN_TYPE) == NODE_UPGRADE:
            return self.nodeAuthNr
        else:
            return super().authNr(req)

    def init_core_authenticator(self):
        return LedgerBasedAuthNr(self.write_manager.txn_types,
                                 self.read_manager.txn_types,
                                 self.action_manager.txn_types,
                                 self.idrCache)

    def defaultNodeAuthNr(self):
        return NodeAuthNr(self.poolLedger)

    async def prod(self, limit: int = None) -> int:
        c = await super().prod(limit)
        c += self.upgrader.service()
        c += self.restarter.service()
        return c

    def can_write_txn(self, txn_type):
        return self.poolCfg.isWritable() or txn_type in [POOL_UPGRADE,
                                                         POOL_CONFIG,
                                                         AUTH_RULE,
                                                         AUTH_RULES]

    def execute_domain_txns(self, three_pc_batch) -> List:
        """
        Execute the REQUEST sent to this Node

        :param ppTime: the time at which PRE-PREPARE was sent
        :param req: the client REQUEST
        """
        return self.default_executer(three_pc_batch)

    def update_txn_with_extra_data(self, txn):
        """
        All the data of the transaction might not be stored in ledger so the
        extra data that is omitted from ledger needs to be fetched from the
        appropriate data store
        :param txn:
        :return:
        """
        # For RAW and ENC attributes, only hash is stored in the ledger.
        if get_type(txn) == ATTRIB:
            txn_data = get_payload_data(txn)
            # The key needs to be present and not None
            key = RAW if (RAW in txn_data and txn_data[RAW] is not None) else \
                ENC if (ENC in txn_data and txn_data[ENC] is not None) else None
            if key:
                txn_data[key] = self.attributeStore.get(txn_data[key])
        return txn

    def is_request_need_quorum(self, msg_dict: dict):
        txn_type = msg_dict.get(OPERATION).get(TXN_TYPE, None) \
            if OPERATION in msg_dict \
            else None
        is_force = OPERATION in msg_dict and msg_dict.get(OPERATION).get(FORCE, False)
        is_force_upgrade = str(is_force) == 'True' and txn_type == POOL_UPGRADE
        return txn_type and not is_force_upgrade and super().is_request_need_quorum(msg_dict)

    @staticmethod
    def add_auth_rules_to_config_state(state: PruningState,
                                       auth_map: dict,
                                       serializer: AbstractConstraintSerializer):
        for rule_id, auth_constraint in auth_map.items():
            serialized_key = rule_id.encode()
            serialized_value = serializer.serialize(auth_constraint)
            if not state.get(serialized_key, isCommitted=False):
                state.set(serialized_key, serialized_value)



import os

from plenum.common.config_helper import PConfigHelper


class ConfigHelper(PConfigHelper):

    @property
    def log_dir(self):
        return self.chroot_if_needed(
            os.path.join(self.config.LOG_DIR, self.config.NETWORK_NAME))

    @property
    def genesis_dir(self):
        return self.chroot_if_needed(
            os.path.join(self.config.GENESIS_DIR, self.config.NETWORK_NAME))

    @property
    def keys_dir(self):
        return self.chroot_if_needed(
            os.path.join(self.config.KEYS_DIR, self.config.NETWORK_NAME, 'keys'))

    @property
    def ledger_base_dir(self):
        return self.chroot_if_needed(self.config.LEDGER_DIR)

    @property
    def ledger_data_dir(self):
        return self.chroot_if_needed(
            os.path.join(self.config.LEDGER_DIR, self.config.NETWORK_NAME, 'data'))

    @property
    def log_base_dir(self):
        return self.chroot_if_needed(self.config.LOG_DIR)

    @property
    def backup_dir(self):
        return self.chroot_if_needed(self.config.BACKUP_DIR)

    @property
    def node_info_dir(self):
        return self.chroot_if_needed(
            os.path.join(self.config.NODE_INFO_DIR, self.config.NETWORK_NAME))


class NodeConfigHelper(ConfigHelper):

    def __init__(self, name: str, config, *, chroot='/'):
        assert name is not None
        super().__init__(config, chroot=chroot)
        self.name = name

    @property
    def ledger_dir(self):
        return self.chroot_if_needed(
            os.path.join(self.config.LEDGER_DIR, self.config.NETWORK_NAME, 'data', self.name))



import pytest
from indy import did

from indy_common.authorize.auth_actions import AuthActionEdit, EDIT_PREFIX, ADD_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_common.authorize.auth_map import promote_node as promote_action
from indy_common.constants import ENDORSER_STRING, NODE, ENDORSER
from indy_node.test.auth_rule.helper import sdk_send_and_check_auth_rule_request, get_pool_validator_count
from plenum.test.pool_transactions.helper import sdk_add_new_nym, demote_node, promote_node


def test_node_txn_promote_by_endorser(txnPoolNodeSet,
                                     sdk_pool_handle,
                                     sdk_wallet_trustee,
                                     looper,
                                     sdk_wallet_handle):

    validators_before = get_pool_validator_count(txnPoolNodeSet)

    new_end_did, new_end_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_trustee[0], "{}"))

    # Step 1. Demote node using default auth rule
    demote_node(looper, sdk_wallet_trustee, sdk_pool_handle, txnPoolNodeSet[-1])

    # Check, that node was demoted
    assert validators_before - get_pool_validator_count(txnPoolNodeSet[:-1]) == 1

    # Step 2. Add new Endorser
    sdk_add_new_nym(looper, sdk_pool_handle,
                    sdk_wallet_trustee,
                    'newEndorser',
                    ENDORSER_STRING,
                    verkey=new_end_verkey, dest=new_end_did)

    new_constraint = AuthConstraint(ENDORSER, 1)

    # Step 3. Change auth rule, to allowing endorser promote node back
    sdk_send_and_check_auth_rule_request(looper, sdk_pool_handle, sdk_wallet_trustee,
                                         auth_action=EDIT_PREFIX,
                                         auth_type=promote_action.txn_type,
                                         field=promote_action.field,
                                         new_value=promote_action.new_value,
                                         old_value=promote_action.old_value,
                                         constraint=new_constraint.as_dict)

    # Step 4. Promote node back, using new Endorser
    promote_node(looper, (sdk_wallet_handle, new_end_did), sdk_pool_handle, txnPoolNodeSet[-1])

    # Check, that all other nodes return previous demoted node back
    assert validators_before == get_pool_validator_count(txnPoolNodeSet[:-1])



#!/usr/bin/python3.5
import os
import sys
import shutil
import traceback
import tarfile
import pwd
import grp
import stat

from stp_core.common.log import getlogger
from storage.kv_store_leveldb import KeyValueStorageLeveldb
from storage.kv_store_rocksdb import KeyValueStorageRocksdb
from storage.kv_store_leveldb_int_keys import KeyValueStorageLeveldbIntKeys
from storage.kv_store_rocksdb_int_keys import KeyValueStorageRocksdbIntKeys

from indy_common.config_util import getConfig
from indy_common.config_helper import NodeConfigHelper


logger = getlogger()

ENV_FILE_PATH = "/etc/indy/indy.env"
ledger_types = ['pool', 'domain', 'config']


def set_own_perm(usr, dir):
    uid = pwd.getpwnam(usr).pw_uid
    gid = grp.getgrnam(usr).gr_gid
    perm_mask_rw = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP
    perm_mask_rwx = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP

    os.chown(dir, uid, gid)
    os.chmod(dir, perm_mask_rwx)
    for croot, sub_dirs, cfiles in os.walk(dir):
        for fs_name in sub_dirs:
            os.chown(os.path.join(croot, fs_name), uid, gid)
            os.chmod(os.path.join(croot, fs_name), perm_mask_rwx)
        for fs_name in cfiles:
            os.chown(os.path.join(croot, fs_name), uid, gid)
            os.chmod(os.path.join(croot, fs_name), perm_mask_rw)


def get_node_name():
    node_name = None
    node_name_key = 'NODE_NAME'

    if os.path.exists(ENV_FILE_PATH):
        with open(ENV_FILE_PATH, "r") as fenv:
            for line in fenv.readlines():
                if line.find(node_name_key) != -1:
                    node_name = line.split('=')[1].strip()
                    break
    else:
        logger.error("Path to env file does not exist")

    return node_name


def archive_leveldb_ledger(node_name, leveldb_ledger_dir):
    leveldb_ledger_archive_name = node_name + "_ledger_leveldb.tar.gz"
    leveldb_ledger_archive_path = os.path.join("/tmp", leveldb_ledger_archive_name)
    tar = tarfile.open(leveldb_ledger_archive_path, "w:gz")
    tar.add(leveldb_ledger_dir, arcname=node_name)
    tar.close()
    logger.info("Archive of LevelDB-based ledger created: {}"
                .format(leveldb_ledger_archive_path))


def migrate_storage(level_db_dir, rocks_db_dir, db_name, is_db_int_keys):
    if is_db_int_keys is True:
        KeyValueStorageLeveldbCls = KeyValueStorageLeveldbIntKeys
        KeyValueStorageRocksdbCls = KeyValueStorageRocksdbIntKeys
    else:
        KeyValueStorageLeveldbCls = KeyValueStorageLeveldb
        KeyValueStorageRocksdbCls = KeyValueStorageRocksdb

    try:
        leveldb_storage = KeyValueStorageLeveldbCls(level_db_dir, db_name, read_only=True)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not open leveldb storage: {}".format(os.path.join(level_db_dir, db_name)))
        return False

    try:
        rocksdb_storage = KeyValueStorageRocksdbCls(rocks_db_dir, db_name)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not open rocksdb storage: {}".format(os.path.join(rocks_db_dir, db_name)))
        return False

    try:
        for key, val in leveldb_storage.iterator():
            rocksdb_storage.put(bytes(key), bytes(val))
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not put key/value to RocksDB storage '{}'".format(db_name))
        return False

    leveldb_storage.close()
    rocksdb_storage.close()

    return True


def migrate_storages(leveldb_ledger_dir, rocksdb_ledger_dir):
    # Migrate transaction logs, they use integer keys
    for ledger_type in ledger_types:
        db_name = ledger_type + "_transactions"
        if not migrate_storage(leveldb_ledger_dir, rocksdb_ledger_dir, db_name, True):
            logger.error("Could not migrate {}, DB path: {}"
                         .format(db_name, os.path.join(leveldb_ledger_dir, db_name)))
            return False

    # Migrate other storages with non-integer keys
    for db_name in ["attr_db", "idr_cache_db", "seq_no_db", "state_signature"]:
        if not migrate_storage(leveldb_ledger_dir, rocksdb_ledger_dir, db_name, False):
            logger.error("Could not migrate {}, DB path: {}"
                         .format(db_name, os.path.join(leveldb_ledger_dir, db_name)))
            return False

    return True


def migrate_all():
    node_name = get_node_name()
    if node_name is None:
        logger.error("Could not get node name")
        return False

    config = getConfig()
    config_helper = NodeConfigHelper(node_name, config)

    leveldb_ledger_dir = config_helper.ledger_dir
    rocksdb_ledger_dir = os.path.join(config_helper.ledger_data_dir, node_name + "_rocksdb")
    if os.path.exists(rocksdb_ledger_dir):
        logger.error("Temporary directory for RocksDB-based ledger exists, please remove: {}"
                     .format(rocksdb_ledger_dir))
        return False

    try:
        os.mkdir(rocksdb_ledger_dir)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not create temporary directory for RocksDB-based ledger: {}"
                     .format(rocksdb_ledger_dir))
        return False

    logger.info("Starting migration of storages from LevelDB to RocksDB...")

    if migrate_storages(leveldb_ledger_dir, rocksdb_ledger_dir):
        logger.info("All storages migrated successfully from LevelDB to RocksDB")
    else:
        logger.error("Storages migration from LevelDB to RocksDB failed!")
        shutil.rmtree(rocksdb_ledger_dir)
        return False

    # Archiving LevelDB-based ledger
    try:
        archive_leveldb_ledger(node_name, leveldb_ledger_dir)
    except Exception:
        logger.warning("Could not create an archive of LevelDB-based ledger, proceed anyway")

    # TODO: it whould be nice to open new RocksDB-based ledger
    # and compare root hashes with LevelDB-based ledger here

    # Remove LevelDB-based ledger
    try:
        shutil.rmtree(leveldb_ledger_dir)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not remove LevelDB-based ledger: {}"
                     .format(leveldb_ledger_dir))
        shutil.rmtree(rocksdb_ledger_dir)
        return False

    ledger_dir = leveldb_ledger_dir

    try:
        shutil.move(rocksdb_ledger_dir, ledger_dir)
    except Exception:
        logger.error(traceback.format_exc())
        logger.error("Could not rename temporary RocksDB-based ledger from '{}' to '{}'"
                     .format(rocksdb_ledger_dir, ledger_dir))
        shutil.rmtree(rocksdb_ledger_dir)
        return False

    set_own_perm("indy", ledger_dir)

    return True


if migrate_all():
    logger.info("Migration from LevelDB to RocksDB complete")
else:
    logger.info("Migration from LevelDB to RocksDB failed!")
    sys.exit(1)



import pytest
from copy import deepcopy

from indy_node.test import waits
from stp_core.loop.eventually import eventually
from plenum.common.constants import VERSION, NAME
from plenum.common.util import randomString
from plenum.test.test_node import checkNodesConnected
from indy_common.constants import SHA256, CANCEL, ACTION
from indy_common.test.conftest import tconf
from indy_node.test.helper import TestNode
from indy_node.test.upgrade.helper import bumpVersion, \
    checkUpgradeScheduled, sdk_ensure_upgrade_sent

whitelist = ['Failed to upgrade node']


@pytest.fixture(scope="module")
def txnPoolNodeSet(tconf, nodeSet):
    for node in nodeSet:
        node.config = tconf
        node.upgrader.config = tconf
    return nodeSet


@pytest.mark.skip(reason='SOV-559')
def testUpgradeLatestUncancelledVersion(looper,
                                        txnPoolNodeSet, tconf, sdk_node_theta_added,
                                        validUpgrade, sdk_pool_handle, sdk_wallet_trustee,
                                        tdirWithPoolTxns, allPluginsPath):
    """
    A node starts and finds several upgrades but selects the latest one which
    is not cancelled, eg node is on version 1.2 but finds 1.3, 1.4 and 1.5 but
    since 1.5 is cancelled, it selects 1.4
    """
    nodeSet = txnPoolNodeSet
    _, newNode = sdk_node_theta_added
    for node in nodeSet[:-1]:
        node.nodestack.removeRemoteByName(newNode.nodestack.name)
        newNode.nodestack.removeRemoteByName(node.nodestack.name)
    newNode.stop()
    nodeSet = nodeSet[:-1]
    looper.removeProdable(newNode)

    upgr1 = deepcopy(validUpgrade)

    upgr2 = deepcopy(upgr1)
    upgr2[VERSION] = bumpVersion(upgr1[VERSION])
    upgr2[NAME] += randomString(3)
    # upgr2[SHA256] = get_valid_code_hash()
    upgr2[SHA256] = 'db34a72a90d026dae49c3b3f0436c8d3963476c77468ad955845a1ccf7b03f55'

    upgr3 = deepcopy(upgr2)
    upgr3[VERSION] = bumpVersion(upgr2[VERSION])
    upgr3[NAME] += randomString(3)
    # upgr3[SHA256] = get_valid_code_hash()
    upgr3[SHA256] = '112c060527e8cecfafe64dcb5bdabc4010cc7b64e0bf9bc2a43d23c37d927128'

    upgr4 = deepcopy(upgr3)
    upgr4[ACTION] = CANCEL

    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgr1)
    looper.run(eventually(checkUpgradeScheduled,
                          nodeSet[:-1],
                          upgr1[VERSION],
                          retryWait=1,
                          timeout=waits.expectedUpgradeScheduled()))

    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgr2)
    looper.run(eventually(checkUpgradeScheduled,
                          nodeSet[:-1],
                          upgr2[VERSION],
                          retryWait=1,
                          timeout=waits.expectedUpgradeScheduled()))

    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgr3)
    looper.run(eventually(checkUpgradeScheduled,
                          nodeSet[:-1],
                          upgr3[VERSION],
                          retryWait=1,
                          timeout=waits.expectedUpgradeScheduled()))

    sdk_ensure_upgrade_sent(looper, sdk_pool_handle, sdk_wallet_trustee, upgr4)
    looper.run(eventually(checkUpgradeScheduled,
                          nodeSet[:-1],
                          upgr2[VERSION],
                          retryWait=1,
                          timeout=waits.expectedUpgradeScheduled()))

    newNode = TestNode(newNode.name, basedirpath=tdirWithPoolTxns, base_data_dir=tdirWithPoolTxns,
                       config=tconf, pluginPaths=allPluginsPath,
                       ha=newNode.nodestack.ha, cliha=newNode.clientstack.ha)
    looper.add(newNode)
    nodeSet.append(newNode)
    looper.run(checkNodesConnected(nodeSet))

    looper.run(eventually(checkUpgradeScheduled, [newNode, ], upgr2[VERSION],
                          retryWait=1, timeout=waits.expectedUpgradeScheduled()))



import json
import time
from indy_common.constants import CRED_DEF_ID, ID, REVOC_TYPE, TAG, GET_REVOC_REG_DEF, VALUE, MAX_CRED_NUM, TXN_TYPE
from indy_common.state import domain
from indy_common.types import Request
from plenum.common.constants import DOMAIN_LEDGER_ID
from plenum.test.helper import sdk_sign_request_from_dict
from plenum.test.helper import sdk_send_and_check


def compare_request_reply(req, reply):
    assert req['operation'][CRED_DEF_ID] == reply['result']['data'][CRED_DEF_ID]
    assert req['operation'][ID] == reply['result']['data'][ID]
    assert req['operation'][TAG] == reply['result']['data'][TAG]
    assert req['operation'][VALUE] == reply['result']['data'][VALUE]


def test_send_get_revoc_reg_def(looper,
                                txnPoolNodeSet,
                                sdk_wallet_steward,
                                sdk_pool_handle,
                                send_revoc_reg_def_by_default):
    _, author_did = sdk_wallet_steward
    revoc_req, _ = send_revoc_reg_def_by_default
    revoc_reg_def_id = revoc_req['operation'][ID]
    get_revoc_reg_def_req = {
        ID: ":".join([author_did,
                      domain.MARKER_REVOC_DEF,
                      revoc_req['operation'][CRED_DEF_ID],
                      revoc_req['operation'][REVOC_TYPE],
                      revoc_req['operation'][TAG]]),
        TXN_TYPE: GET_REVOC_REG_DEF,
    }
    get_revoc_reg_def_req = sdk_sign_request_from_dict(looper,
                                                       sdk_wallet_steward,
                                                       get_revoc_reg_def_req)
    replies = sdk_send_and_check([json.dumps(get_revoc_reg_def_req)],
                                 looper,
                                 txnPoolNodeSet,
                                 sdk_pool_handle)
    req, reply = replies[0]
    compare_request_reply(revoc_req, reply)


def test_get_revoc_reg_def_from_uncommited(looper,
                                           txnPoolNodeSet,
                                           sdk_wallet_steward,
                                           sdk_pool_handle,
                                           send_revoc_reg_def_by_default):
    # REVOC_REG_DEF was added into pool by send_revoc_reg_def_by_default fixture

    _, author_did = sdk_wallet_steward
    revoc_req, _ = send_revoc_reg_def_by_default
    new_maxCredNum = 100
    revoc_req['operation'][VALUE][MAX_CRED_NUM] = new_maxCredNum
    revoc_req = sdk_sign_request_from_dict(looper, sdk_wallet_steward, revoc_req['operation'])

    # We apply transacttion, which will be not commited

    for node in txnPoolNodeSet:
        node.write_manager.apply_request(Request(**revoc_req), int(time.time()))
    get_revoc_reg_def_req = {
        ID: ":".join([author_did,
                      domain.MARKER_REVOC_DEF,
                      revoc_req['operation'][CRED_DEF_ID],
                      revoc_req['operation'][REVOC_TYPE],
                      revoc_req['operation'][TAG]]),
        TXN_TYPE: GET_REVOC_REG_DEF,
    }
    get_revoc_reg_def_req = sdk_sign_request_from_dict(looper,
                                                       sdk_wallet_steward,
                                                       get_revoc_reg_def_req)

    # Send GET_REVOC_REG_DEF query.
    # We expects that commited REVOC_REG_DEF transaction will be returned

    replies = sdk_send_and_check([json.dumps(get_revoc_reg_def_req)],
                                 looper,
                                 txnPoolNodeSet,
                                 sdk_pool_handle)
    req, reply = replies[0]
    assert new_maxCredNum != reply['result']['data'][VALUE][MAX_CRED_NUM]



from abc import ABCMeta, abstractmethod
from typing import Optional

from common.serializers.json_serializer import JsonSerializer
from indy_common.authorize.auth_actions import AuthActionEdit, AuthActionAdd
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.constants import DOMAIN_LEDGER_ID, RS_ID, RS_TYPE, RS_VERSION, RS_NAME, RS_CONTENT, JSON_LD_ID_FIELD, \
    JSON_LD_TYPE_FIELD
from indy_common.state.domain import encode_state_value
from indy_common.types import Request
from indy_common.config_util import getConfig
from plenum.common.constants import TXN_PAYLOAD_METADATA_ENDORSER, TXN_PAYLOAD_METADATA_FROM, TXN_PAYLOAD_VERSION
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.txn_util import get_payload_data, get_seq_no, get_txn_time, get_from, get_endorser, \
    get_payload_txn_version
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.write_request_handler import WriteRequestHandler


class AbstractRichSchemaObjectHandler(WriteRequestHandler, metaclass=ABCMeta):

    def __init__(self, txn_type, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(database_manager, txn_type, DOMAIN_LEDGER_ID)
        self.write_req_validator = write_req_validator
        self.config = getConfig()

    def _enabled(self) -> bool:
        return self.config.ENABLE_RICH_SCHEMAS

    def _validate_enabled(self, request: Request):
        if not self._enabled():
            raise InvalidClientRequest(request.identifier, request.reqId, "RichSchema transactions are disabled")

    @abstractmethod
    def is_json_ld_content(self):
        pass

    @abstractmethod
    def do_static_validation_content(self, content_as_dict, request):
        pass

    @abstractmethod
    def do_dynamic_validation_content(self, request):
        pass

    def static_validation(self, request: Request):
        self._validate_request_type(request)
        self._validate_enabled(request)

        try:
            content_as_dict = JsonSerializer.loads(request.operation[RS_CONTENT])
        except ValueError:
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "'{}' must be a JSON serialized string".format(RS_CONTENT))

        if self.is_json_ld_content():
            self.do_static_validation_json_ld(content_as_dict, request)

        self.do_static_validation_content(content_as_dict, request)

    def do_static_validation_json_ld(self, content_as_dict, request):
        if not content_as_dict.get(JSON_LD_ID_FIELD):
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "'content' must be a valid JSON-LD and have non-empty '{}' field".format(JSON_LD_ID_FIELD))
        if not content_as_dict.get(JSON_LD_TYPE_FIELD):
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "'content' must be a valid JSON-LD and have non-empty '{}' field".format(
                                           JSON_LD_TYPE_FIELD))

        if content_as_dict[JSON_LD_ID_FIELD] != request.operation[RS_ID]:
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "content's @id must be equal to id={}".format(request.operation[RS_ID]))

    def additional_dynamic_validation(self, request: Request, req_pp_time: Optional[int]):
        self._validate_request_type(request)

        rs_id = request.operation[RS_ID]
        rs_object, _, _ = self.get_from_state(rs_id)

        # check that (rs_name, rs_type, rs_version) is unique within all rich schema objects
        secondary_key = self.make_secondary_key(request.operation[RS_TYPE],
                                                request.operation[RS_NAME],
                                                request.operation[RS_VERSION])
        if not rs_object and self.state.get(secondary_key, isCommitted=False) is not None:
            raise InvalidClientRequest(request.identifier,
                                       request.reqId,
                                       'An object with {rs_name}="{rs_name_value}", {rs_version}="{rs_version_value}" '
                                       'and {rs_type}="{rs_type_value}" already exists. '
                                       'Please choose different {rs_name}, {rs_version} or {rs_type}'.format(
                                           rs_name=RS_NAME, rs_version=RS_VERSION, rs_type=RS_TYPE,
                                           rs_name_value=request.operation[RS_NAME],
                                           rs_version_value=request.operation[RS_VERSION],
                                           rs_type_value=request.operation[RS_TYPE]))

        # do common auth-rule-based validation (which will check the default immutability of most of the objects)
        if rs_object:
            self.write_req_validator.validate(request,
                                              [AuthActionEdit(txn_type=self.txn_type,
                                                              field='*',
                                                              old_value='*',
                                                              new_value='*')])
        else:
            self.write_req_validator.validate(request,
                                              [AuthActionAdd(txn_type=self.txn_type,
                                                             field='*',
                                                             value='*')])

        self.do_dynamic_validation_content(request)

    def update_state(self, txn, prev_result, request, is_committed=False) -> None:
        self._validate_txn_type(txn)

        txn_data = get_payload_data(txn)

        primary_key = txn_data[RS_ID].encode()
        secondary_key = self.make_secondary_key(txn_data[RS_TYPE],
                                                txn_data[RS_NAME],
                                                txn_data[RS_VERSION])

        value = {
            RS_ID: txn_data[RS_ID],
            RS_TYPE: txn_data[RS_TYPE],
            RS_NAME: txn_data[RS_NAME],
            RS_VERSION: txn_data[RS_VERSION],
            RS_CONTENT: txn_data[RS_CONTENT],
            TXN_PAYLOAD_METADATA_FROM: get_from(txn),
            TXN_PAYLOAD_METADATA_ENDORSER: get_endorser(txn),
            TXN_PAYLOAD_VERSION: get_payload_txn_version(txn),
        }
        seq_no = get_seq_no(txn)
        txn_time = get_txn_time(txn)
        value_bytes = encode_state_value(value, seq_no, txn_time)

        self.state.set(primary_key, value_bytes)
        self.state.set(secondary_key, primary_key)

    def gen_txn_id(self, txn):
        self._validate_txn_type(txn)
        return get_payload_data(txn)[RS_ID]

    @staticmethod
    def make_secondary_key(rs_type, rs_name, rs_version):
        return "{RS_TYPE}:{RS_NAME}:{RS_VERSION}".format(RS_TYPE=rs_type,
                                                         RS_NAME=rs_name,
                                                         RS_VERSION=rs_version).encode()



from indy_common.authorize.auth_actions import AuthActionAdd
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.constants import VALIDATOR_INFO
from indy_node.server.request_handlers.action_req_handlers.utils import generate_action_result
from plenum.common.constants import DATA
from plenum.common.request import Request
from plenum.common.txn_util import get_request_data
from plenum.server.database_manager import DatabaseManager

from plenum.server.request_handlers.handler_interfaces.action_request_handler import ActionRequestHandler
from stp_core.common.log import getlogger

logger = getlogger()


class ValidatorInfoHandler(ActionRequestHandler):
    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator,
                 info_tool):
        super().__init__(database_manager, VALIDATOR_INFO, None)
        self.write_req_validator = write_req_validator
        self.info_tool = info_tool

    def static_validation(self, request: Request):
        pass

    def dynamic_validation(self, request: Request):
        self._validate_request_type(request)
        self.write_req_validator.validate(request,
                                          [AuthActionAdd(txn_type=VALIDATOR_INFO,
                                                         field='*',
                                                         value='*')])

    def process_action(self, request: Request):
        self._validate_request_type(request)
        identifier, req_id, operation = get_request_data(request)
        logger.debug("Transaction {} with type {} started"
                     .format(req_id, request.txn_type))
        result = generate_action_result(request)
        result[DATA] = self.info_tool.info
        # Memory profiler needs only for debug purposes,
        # because collecting memory info it's a very long time operation
        # result[DATA].update(self.info_tool.memory_profiler)
        """For keeping output format"""
        result[DATA].update({'Memory_profiler': []})
        result[DATA].update(self.info_tool._generate_software_info())
        result[DATA].update(self.info_tool.extractions)
        # ToDo: also can be a long time operation, because we use external tools
        result[DATA].update(self.info_tool.node_disk_size)
        logger.debug("Transaction {} with type {} finished"
                     .format(req_id, request.txn_type))
        return result



import pytest

from common.serializers.serialization import domain_state_serializer
from indy_common.authorize.auth_cons_strategies import ConfigLedgerAuthStrategy
from indy_common.authorize.auth_constraints import AuthConstraint, ConstraintsSerializer
from indy_common.state import config
from plenum.common.constants import TRUSTEE
from state.pruning_state import PruningState
from storage.kv_in_memory import KeyValueStorageInMemory


@pytest.fixture
def state():
    return PruningState(KeyValueStorageInMemory())


@pytest.fixture
def state_serializer():
    return ConstraintsSerializer(domain_state_serializer)


def test_config_strategy_get_constraint_from_state(state,
                                                   state_serializer):
    action_id = "1--2--3--4--5"
    constraint_to_state = AuthConstraint(role=TRUSTEE,
                                         sig_count=1,
                                         need_to_be_owner=True)
    constraint_to_map = AuthConstraint(role=TRUSTEE,
                                       sig_count=5,
                                       need_to_be_owner=False)
    auth_map = {action_id: constraint_to_map}
    state.set(config.make_state_path_for_auth_rule(action_id),
              state_serializer.serialize(constraint_to_state))
    strategy = ConfigLedgerAuthStrategy(auth_map=auth_map,
                                        state=state,
                                        serializer=state_serializer)
    from_state = strategy.get_auth_constraint(action_id)
    assert from_state != constraint_to_map
    assert from_state == constraint_to_state


def test_config_strategy_constraint_from_map(state,
                                             state_serializer):
    action_id = "1--2--3--4--5"
    constraint = AuthConstraint(role=TRUSTEE,
                                sig_count=1,
                                need_to_be_owner=True)
    auth_map = {action_id: constraint}
    strategy = ConfigLedgerAuthStrategy(auth_map=auth_map,
                                        state=state,
                                        serializer=state_serializer)
    from_map = strategy.get_auth_constraint(action_id)
    assert from_map == constraint


def test_config_strategy_constraint_not_found(state,
                                              state_serializer):
    state_action_id = "1--2--3--4--5"
    test_action_id = "1--2--3--4--50"
    constraint = AuthConstraint(role=TRUSTEE,
                                sig_count=1,
                                need_to_be_owner=True)
    auth_map = {state_action_id: constraint}
    state.set(config.make_state_path_for_auth_rule(state_action_id), state_serializer.serialize(constraint))
    strategy = ConfigLedgerAuthStrategy(auth_map=auth_map,
                                        state=state,
                                        serializer=state_serializer)
    from_state = strategy.get_auth_constraint(test_action_id)
    assert from_state is None



import pytest
from indy import did

from indy_common.constants import NETWORK_MONITOR
from indy_node.test.validator_info.helper import sdk_get_validator_info
from plenum.common.constants import STEWARD_STRING
from plenum.common.exceptions import RequestRejectedException
from plenum.test.helper import sdk_sign_and_submit_op, sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_add_new_nym


def test_network_monitor_suspension_by_another_steward(looper,
                                                       sdk_pool_handle,
                                                       sdk_wallet_steward,
                                                       sdk_wallet_trustee,
                                                       sdk_wallet_handle,
                                                       with_verkey):
    new_steward_did, new_steward_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_trustee[0], "{}"))
    new_network_monitor_did, new_network_monitor_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_steward[0], "{}"))

    """Adding new steward"""
    sdk_add_new_nym(looper, sdk_pool_handle,
                    sdk_wallet_trustee, 'newSteward', STEWARD_STRING, verkey=new_steward_verkey, dest=new_steward_did)

    """Adding NETWORK_MONITOR role by first steward"""
    op = {'type': '1',
          'dest': new_network_monitor_did,
          'role': NETWORK_MONITOR,
          'verkey': new_network_monitor_verkey}
    req = sdk_sign_and_submit_op(looper, sdk_pool_handle, (sdk_wallet_handle, new_steward_did), op)
    sdk_get_and_check_replies(looper, [req])

    """Check that get_validator_info command works for NETWORK_MONITOR role"""
    sdk_get_validator_info(looper, (sdk_wallet_handle, new_network_monitor_did), sdk_pool_handle)

    """Blacklisting network_monitor by new steward"""
    op = {'type': '1',
          'dest': new_network_monitor_did,
          'role': None}
    if with_verkey:
        op['verkey'] = new_network_monitor_verkey
    req = sdk_sign_and_submit_op(looper, sdk_pool_handle, (sdk_wallet_handle, new_steward_did), op)
    if with_verkey:
        with pytest.raises(RequestRejectedException):
            sdk_get_and_check_replies(looper, [req])
    else:
        sdk_get_and_check_replies(looper, [req])
        with pytest.raises(RequestRejectedException):
            sdk_get_validator_info(looper, (sdk_wallet_handle, new_network_monitor_did), sdk_pool_handle)


def test_network_monitor_suspension_by_itself(looper,
                                              sdk_pool_handle,
                                              sdk_wallet_steward,
                                              sdk_wallet_handle,
                                              with_verkey):
    new_network_monitor_did, new_network_monitor_verkey = looper.loop.run_until_complete(
        did.create_and_store_my_did(sdk_wallet_steward[0], "{}"))

    """Adding NETWORK_MONITOR role by steward"""
    op = {'type': '1',
          'dest': new_network_monitor_did,
          'role': NETWORK_MONITOR,
          'verkey': new_network_monitor_verkey}
    req = sdk_sign_and_submit_op(looper, sdk_pool_handle, (sdk_wallet_handle, sdk_wallet_steward[1]), op)
    sdk_get_and_check_replies(looper, [req])

    """Blacklisting network_monitor by itself"""
    op = {'type': '1',
          'dest': new_network_monitor_did,
          'role': None}
    if with_verkey:
        op['verkey'] = new_network_monitor_verkey
    req = sdk_sign_and_submit_op(looper, sdk_pool_handle, (sdk_wallet_handle, new_network_monitor_did), op)
    with pytest.raises(RequestRejectedException):
        sdk_get_and_check_replies(looper, [req])



from indy_node.test import waits
from plenum.test.pool_transactions.helper import demote_node
from stp_core.loop.eventually import eventually
from plenum.common.constants import VERSION
from indy_common.constants import SCHEDULE
from indy_node.test.upgrade.helper import checkUpgradeScheduled, sdk_ensure_upgrade_sent


def test_update_with_demoted_node(looper, nodeSet, validUpgrade,
                                  sdk_pool_handle, sdk_wallet_stewards,
                                  sdk_wallet_trustee):
    # demote one node
    demote_node(looper, sdk_wallet_stewards[3], sdk_pool_handle, nodeSet[3])

    # remove demoted node from upgrade schedule
    upgr = validUpgrade
    del upgr[SCHEDULE][nodeSet[3].id]

    # send upgrade
    sdk_ensure_upgrade_sent(looper, sdk_pool_handle,
                            sdk_wallet_trustee, upgr)

    # check upg scheduled
    looper.run(eventually(checkUpgradeScheduled, nodeSet[:3], upgr[VERSION], retryWait=1,
                          timeout=waits.expectedUpgradeScheduled()))



from indy_common.authorize.auth_actions import AuthActionEdit
from indy_common.constants import POOL_CONFIG, ACTION, AUTH_RULE


def test_auth_rule(write_request_validation, is_owner, req):
    authorized = req.identifier == "trustee_identifier"
    assert authorized == write_request_validation(req,
                                                  [AuthActionEdit(txn_type=AUTH_RULE,
                                                                  field=ACTION,  # TODO: change to real field name
                                                                  old_value='old_value',
                                                                  new_value='new_value',
                                                                  is_owner=is_owner)])



import pytest
from plenum.test.pool_transactions.helper import sdk_add_new_nym

from indy_node.test.auth_rule.helper import sdk_send_and_check_auth_rule_request
from plenum.test.delayers import req_delay


@pytest.fixture(scope="module")
def nodeSetWithOneNodeResponding(nodeSet):
    # the order of nodes the client sends requests to is [Alpha, Beta, Gamma, Delta]
    # delay all requests to Beta, Gamma and Delta
    # we expect that it's sufficient for the client to get Reply from Alpha only
    # as for write requests, we can send it to 1 node only, and it will be propagated to others
    for node in nodeSet[1:]:
        node.clientIbStasher.delay(req_delay())
    return nodeSet


@pytest.fixture(scope="module")
def send_auth_rule(looper, sdk_pool_handle, sdk_wallet_trustee, nodeSet):
    return sdk_send_and_check_auth_rule_request(looper,
                                                sdk_pool_handle,
                                                sdk_wallet_trustee)



import pytest
import os

from indy_node.server.restart_log import RestartLog, RestartLogData
import datetime

@pytest.fixture
def log_file_path(tdir, request):
    return os.path.join(
        tdir,
        "{}.restart_log".format(os.path.basename(request.node.nodeid))
    )


def test_restart_log_data_pack_unpack():
    delimiter = '|'
    data = RestartLogData(datetime.datetime.utcnow())
    assert data == RestartLogData.unpack(
        data.pack(delimiter=delimiter), delimiter=delimiter
    )



# TODO actually it is already well tested in base calss ActionLog
@pytest.mark.parametrize('ev_type', RestartLog.Events)
def test_restart_log_append_api(log_file_path, ev_type):
    restart_log = RestartLog(log_file_path)
    ev_data = RestartLogData(datetime.datetime.utcnow())
    getattr(restart_log, "append_{}".format(ev_type.name))(ev_data)
    assert restart_log.last_event.data == ev_data
    assert restart_log.last_event.ev_type == ev_type


def test_restart_log_loads_legacy_data(monkeypatch, log_file_path):

    ev_index = None
    tss = [
        '2019-02-28 07:36:23.135789',
        '2019-02-28 07:37:11.008484',
        '2019-02-28 07:38:33.721644'
    ]

    class datetime_wrapper(datetime.datetime):
        def utcnow():
            return tss[ev_index]

    monkeypatch.setattr(datetime, 'datetime', datetime_wrapper)

    legacy_logs = (
        "{}\tscheduled\t2019-02-28 07:37:11+00:00\r\n".format(tss[0]) +
        "{}\tstarted\t2019-02-28 07:37:11+00:00\r\n".format(tss[1]) +
        "{}\tsucceeded\t2019-02-28 07:37:11+00:00\r\n".format(tss[2])
    )

    with open(log_file_path, 'w', newline='') as f:
        f.write(legacy_logs)
    restart_log_legacy = RestartLog(log_file_path)

    log_file_path_new = log_file_path + '_new'
    restart_log_new = RestartLog(log_file_path_new)

    for ev_index, ev in enumerate(restart_log_legacy):
        getattr(restart_log_new, 'append_' + ev.ev_type.name)(ev.data)

    with open(log_file_path_new, 'r', newline='') as f:
        new_logs = f.read()

    assert legacy_logs == new_logs



import pytest

from indy_common.constants import FAIL
from indy_node.test.upgrade.helper import populate_log_with_upgrade_events, \
    bumpedVersion, check_node_sent_acknowledges_upgrade, check_node_do_not_sent_acknowledges_upgrade, \
    emulate_restart_pool_for_upgrade
from indy_node.server.upgrade_log import UpgradeLog

INVALID_VERSION = bumpedVersion()
whitelist = ['unable to send message',
             'failed upgrade',
             'This problem may have external reasons, check syslog for more information']


# TODO: Implement a client in node


@pytest.fixture(scope="module")
def tdirWithPoolTxns(tdirWithPoolTxns, poolTxnNodeNames, tdir, tconf):
    # For each node, adding a file with he current version number which makes the node
    # think that an upgrade has been performed
    populate_log_with_upgrade_events(
        poolTxnNodeNames, tdir, tconf, INVALID_VERSION)
    return tdirWithPoolTxns


def test_node_detected_upgrade_failed(nodeSet):
    '''
    Test that each node checks Upgrade Log on startup (after Upgrade restart), and writes FAIL to it
    because the current version differs from the one in Upgrade log.
    Upgrade log already created START event (see fixture above emulating real upgrade)
    '''
    for node in nodeSet:
        assert node.upgrader.scheduledAction is None
        assert node.upgrader.lastActionEventInfo.ev_type == UpgradeLog.Events.failed


def test_node_sent_upgrade_fail(looper, nodeSet, nodeIds):
    '''
    Test that each node sends NODE_UPGRADE Fail event (because it sees FAIL in Upgrade log)
    Upgrade log already created START event (see fixture above emulating real upgrade)
    '''
    check_node_sent_acknowledges_upgrade(looper, nodeSet, nodeIds,
                                         allowed_actions=[FAIL],
                                         ledger_size=len(nodeSet),
                                         expected_version=INVALID_VERSION)


def test_node_sent_upgrade_unsuccessful_once(looper, nodeSet, nodeIds):
    '''
    Test that each node sends NODE_UPGRADE Fail event only once,
    so that if we restart the node it's not sent again
    '''
    # emulate restart
    emulate_restart_pool_for_upgrade(nodeSet)

    # check that config ledger didn't changed (no new txns were sent)
    check_node_do_not_sent_acknowledges_upgrade(looper, nodeSet, nodeIds,
                                                allowed_actions=[FAIL],
                                                ledger_size=len(nodeSet),
                                                expected_version=INVALID_VERSION)



import json

import pytest
from indy.ledger import build_acceptance_mechanisms_request

from plenum.common.util import randomString
from plenum.test.helper import sdk_get_and_check_replies
from plenum.test.pool_transactions.helper import sdk_sign_and_send_prepared_request


@pytest.fixture(scope="module")
def setup_aml(looper, txnPoolNodeSet, taa_aml_request_module, sdk_pool_handle, sdk_wallet_trustee):
    req = sdk_sign_and_send_prepared_request(looper, sdk_wallet_trustee, sdk_pool_handle, taa_aml_request_module)
    sdk_get_and_check_replies(looper, [req])


@pytest.fixture(scope="module")
def taa_aml_request_module(looper, sdk_pool_handle, sdk_wallet_trustee):
    return looper.loop.run_until_complete(build_acceptance_mechanisms_request(
        sdk_wallet_trustee[1],
        json.dumps({
            'Nice way': 'very good way to accept agreement'}),
        randomString(), randomString()))



import functools
import time

import pytest
from orderedset._orderedset import OrderedSet

from indy_common.constants import CONFIG_LEDGER_ID
from indy_common.state import config
from indy_node.test.request_handlers.test_update_state_config_req_handler import prepare_request
from plenum.server.consensus.ordering_service import OrderingService
from plenum.test.testing_utils import FakeSomething


@pytest.fixture
def fake_ordering_service(config_ledger,
                 config_state,
                 db_manager):
    ordering_service = FakeSomething(db_manager=db_manager,
                                     post_batch_rejection=lambda *args, **kwargs: True,
                                     _logger=FakeSomething(info=lambda *args, **kwargs: True),
                                     requestQueues={0: OrderedSet(), 1: OrderedSet(), 2: OrderedSet()})
    ordering_service._revert = functools.partial(OrderingService._revert, ordering_service)
    return ordering_service


def test_revert_uncommitted_state(write_manager,
                                  db_manager,
                                  constraint_serializer,
                                  prepare_request,
                                  fake_ordering_service,
                                  tmpdir_factory):
    action, constraint, request = prepare_request
    req_count = 1

    state_root_hash_before = db_manager.get_state(CONFIG_LEDGER_ID).headHash

    write_manager.apply_request(request, int(time.time()))
    """
    Check, that request exist in uncommitted state
    """
    from_state = db_manager.get_state(CONFIG_LEDGER_ID).get(
        config.make_state_path_for_auth_rule(action.get_action_id()),
        isCommitted=False)
    assert constraint_serializer.deserialize(from_state) == constraint

    fake_ordering_service._revert(CONFIG_LEDGER_ID,
                                   state_root_hash_before,
                                   req_count)
    """
    Txn is reverted from ledger and state
    """
    assert len(db_manager.get_ledger(CONFIG_LEDGER_ID).uncommittedTxns) == 0
    from_state = db_manager.get_state(CONFIG_LEDGER_ID).get(
        config.make_state_path_for_auth_rule(action.get_action_id()),
        isCommitted=False)
    assert from_state is None



import pytest
from indy_node.utils.node_control_utils import NodeControlUtil


@pytest.mark.parametrize("vers_str,vers_parsed",
                         [("aa (= 1)", ["aa=1"]), ("aa (= 1), bb", ["aa=1", "bb"]),
                          ("aa (= 1), bb (= 2) | cc (= 3)", ["aa=1", "bb=2", "cc=3"]),
                          ("aa (< 1), bb (> 2) | cc (<= 3), dd (>= 4)", ["aa=1", "bb=2", "cc=3", "dd=4"]),
                          ("aa (<< 1), bb (>> 2) | cc (<= 3) | dd", ["aa=1", "bb=2", "cc=3", "dd"])])
def test_version_parse(vers_str, vers_parsed):
    vers = NodeControlUtil._parse_deps(vers_str)
    assert vers == vers_parsed


@pytest.mark.parametrize("pkgs,pkgs_dd",
                         [([], []), (["aa=1"], ["aa=1"]), (["aa=1", "aa=1", "aa=2"], ["aa=1"]),
                          (["aa=1", "bb=2", "cc=3"], ["aa=1", "bb=2", "cc=3"]),
                          (["aa=1", "bb=2", "cc=3", "aa=2", "bb=3", "cc=4"], ["aa=1", "bb=2", "cc=3"])])
def test_pkgs_dedup(pkgs, pkgs_dd):
    processed_pkgs = NodeControlUtil._pkgs_dedup(pkgs)
    assert processed_pkgs == pkgs_dd



import dateutil

from datetime import datetime, timedelta

from indy_common.constants import START
from indy_node.server.restart_log import RestartLog
from indy_node.test.pool_restart.helper import _createServer, sdk_send_restart, _stopServer
from indy_node.test.pool_restart.test_pool_restart import _comparison_reply, _check_restart_log


def test_pool_restarts_one_by_one_with_restart_now(
        sdk_pool_handle, sdk_wallet_trustee, looper, tconf, txnPoolNodeSet):
    server, indicator = looper.loop.run_until_complete(
        _createServer(
            host=tconf.controlServiceHost,
            port=tconf.controlServicePort
        )
    )

    unow = datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
    first_start = str(datetime.isoformat(unow + timedelta(seconds=1000)))
    req_obj, resp = sdk_send_restart(looper,
                                     sdk_wallet_trustee,
                                     sdk_pool_handle,
                                     action=START,
                                     datetime=first_start)
    _comparison_reply(resp, req_obj)
    second_start = "0"
    req_obj, resp = sdk_send_restart(looper,
                                     sdk_wallet_trustee,
                                     sdk_pool_handle,
                                     action=START,
                                     datetime=second_start)
    _comparison_reply(resp, req_obj)
    restart_log = []
    for a in txnPoolNodeSet[0].restarter._actionLog:
        restart_log.append(a)
    restart_log.reverse()
    _check_restart_log(restart_log[1], RestartLog.Events.scheduled, first_start)
    _check_restart_log(restart_log[0], RestartLog.Events.cancelled)
    _stopServer(server)



import pytest
from indy_common.constants import ENDORSER_STRING

from plenum.common.exceptions import RequestRejectedException

from plenum.test.pool_transactions.helper import sdk_add_new_nym


def testStewardCreatesAEndorser(looper, sdk_pool_handle, sdk_wallet_steward):
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_steward, role=ENDORSER_STRING)


# FIXME why is it necessary to check
def testStewardCreatesAnotherEndorser(looper, sdk_pool_handle, sdk_wallet_steward):
    sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_steward, role=ENDORSER_STRING)



import json

import pytest
from indy.anoncreds import issuer_create_schema, issuer_create_and_store_credential_def
from indy.ledger import build_attrib_request, sign_request, build_schema_request, build_cred_def_request, \
    build_nym_request, build_get_schema_request, parse_get_schema_response

from indy_node.test.helper import createHalfKeyIdentifierAndAbbrevVerkey
from indy_common.types import Request
from indy_node.test.api.helper import sdk_write_schema, build_rs_schema_request
from plenum.common.messages.node_messages import Propagate
from plenum.common.types import f
from plenum.server.message_handlers import PropagateHandler
from plenum.test.helper import sdk_sign_and_submit_req, sdk_get_reply


@pytest.fixture(scope="module")
def node(nodeSet):
    return nodeSet[0]


def emulate_received(node, msg):
    return node.nodestack.deserializeMsg(
        node.nodestack.sign_and_serialize(msg)
    )


@pytest.fixture(scope="module", params=["ATTRIB", "SCHEMA", "RS_SCHEMA", "CLAIM_DEF", "NYM"])
def req(request, looper, sdk_pool_handle, sdk_wallet_steward):
    wallet_handle, identifier = sdk_wallet_steward
    if request.param == "ATTRIB":
        raw = json.dumps({'answer': 42})
        request_json = looper.loop.run_until_complete(
            build_attrib_request(identifier, identifier, raw=raw, xhash=None, enc=None))
    elif request.param == "SCHEMA":
        _, schema_json = looper.loop.run_until_complete(
            issuer_create_schema(identifier, "name", "1.0", json.dumps(["first", "last"])))
        request_json = looper.loop.run_until_complete(build_schema_request(identifier, schema_json))
    elif request.param == "RS_SCHEMA":
        rs_schema = {'@id': "fakeId234e", '@type': "0od"}
        request_json = build_rs_schema_request(identifier, rs_schema, "ISO18023_Drivers_License", "1.1")
    elif request.param == "CLAIM_DEF":
        schema_json, _ = sdk_write_schema(looper, sdk_pool_handle, sdk_wallet_steward)
        schema_id = json.loads(schema_json)['id']

        request = looper.loop.run_until_complete(build_get_schema_request(identifier, schema_id))
        reply = sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_steward, request))[1]
        _, schema_json = looper.loop.run_until_complete(parse_get_schema_response(json.dumps(reply)))

        _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
            wallet_handle, identifier, schema_json, "some_tag", "CL", json.dumps({"support_revocation": True})))
        request_json = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    elif request.param == "NYM":
        idr, verkey = createHalfKeyIdentifierAndAbbrevVerkey()
        request_json = looper.loop.run_until_complete(build_nym_request(identifier, idr, verkey, None, None))

    req_signed = looper.loop.run_until_complete(sign_request(wallet_handle, identifier, request_json))
    return Request(**json.loads(req_signed))


def test_propagate_request(looper, node, sdk_wallet_steward, req):
    # Emulate Received PROPAGATE with the req
    propagate = Propagate(req.as_dict, "client_name")
    msg = emulate_received(node, propagate)

    # Create a new (requested) Propagate from the received data
    ph = PropagateHandler(node)
    kwargs = {f.DIGEST.nm: req.digest}
    received_propagate = ph.create(msg, **kwargs)
    assert received_propagate == propagate



import dateutil

from indy_common.constants import APP_NAME
from indy_node.server.upgrade_log import UpgradeLogData, UpgradeLog
from indy_node.test.upgrade.helper import emulate_view_change_pool_for_upgrade, count_action_log_package

whitelist = ['unable to send message']


# TODO: Implement a client in node


def test_scheduled_once_after_view_change(nodeSet, validUpgrade, upgradeScheduled):
    '''
    Test that each node schedules update only once after each view change
    '''
    # emulate view changes 1-4
    emulate_view_change_pool_for_upgrade(nodeSet)
    emulate_view_change_pool_for_upgrade(nodeSet)
    emulate_view_change_pool_for_upgrade(nodeSet)
    emulate_view_change_pool_for_upgrade(nodeSet)

    # check that there are no cancel events in Upgrade log
    version = validUpgrade['version']
    upgrade_id = nodeSet[0].upgrader.scheduledAction.upgrade_id
    for node in nodeSet:
        node_id = node.poolManager.get_nym_by_name(node.name)
        when = dateutil.parser.parse(validUpgrade['schedule'][node_id])
        ev_data = UpgradeLogData(when, version, upgrade_id, validUpgrade['package'])
        assert node.upgrader.scheduledAction == ev_data
        assert count_action_log_package(list(node.upgrader._actionLog), validUpgrade['package']) == 1
        assert node.upgrader.lastActionEventInfo.ev_type == UpgradeLog.Events.scheduled
        assert node.upgrader.lastActionEventInfo.data == ev_data



import json

from indy.anoncreds import issuer_create_and_store_credential_def
from indy.ledger import build_cred_def_request, build_get_schema_request, parse_get_schema_response
from indy_node.test.api.helper import validate_write_reply, validate_claim_def_txn, sdk_write_schema
from plenum.test.helper import sdk_get_reply, sdk_sign_and_submit_req


def test_claim_def_reply_is_valid(looper, sdk_pool_handle, sdk_wallet_steward):
    wallet_handle, identifier = sdk_wallet_steward

    schema_json, _ = sdk_write_schema(looper, sdk_pool_handle, sdk_wallet_steward)
    schema_id = json.loads(schema_json)['id']

    request = looper.loop.run_until_complete(build_get_schema_request(identifier, schema_id))
    reply = sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_steward, request))[1]
    _, schema_json = looper.loop.run_until_complete(parse_get_schema_response(json.dumps(reply)))

    _, definition_json = looper.loop.run_until_complete(issuer_create_and_store_credential_def(
        wallet_handle, identifier, schema_json, "some_tag", "CL", json.dumps({"support_revocation": True})))
    request = looper.loop.run_until_complete(build_cred_def_request(identifier, definition_json))
    reply = sdk_get_reply(looper, sdk_sign_and_submit_req(sdk_pool_handle, sdk_wallet_steward, request))[1]

    validate_write_reply(reply)
    validate_claim_def_txn(reply['result']['txn'])



import json
import logging
import time
import warnings
import pytest

from common.serializers import serialization
from common.serializers.serialization import domain_state_serializer
from indy_common.authorize.auth_constraints import ConstraintsSerializer
from indy_common.authorize.auth_map import auth_map
from indy_common.authorize.auth_request_validator import WriteRequestValidator
from indy_common.test.constants import IDENTIFIERS
from indy_node.persistence.attribute_store import AttributeStore
from indy_node.persistence.idr_cache import IdrCache
from indy_node.server.node import Node
from indy_node.server.node_bootstrap import NodeBootstrap
from ledger.compact_merkle_tree import CompactMerkleTree
from plenum.bls.bls_store import BlsStore
from plenum.common.exceptions import UnauthorizedClientRequest
from plenum.common.ledger import Ledger
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_managers.action_request_manager import ActionRequestManager
from plenum.server.request_managers.read_request_manager import ReadRequestManager
from plenum.server.request_managers.write_request_manager import WriteRequestManager
from plenum.test.pool_transactions.helper import sdk_add_new_nym, sdk_pool_refresh, prepare_new_node_data, \
    create_and_start_new_node, prepare_node_request, sdk_sign_and_send_prepared_request, prepare_nym_request
from plenum.test.testing_utils import FakeSomething
from state.pruning_state import PruningState
from storage.kv_in_memory import KeyValueStorageInMemory
from stp_core.common.log import Logger

from plenum.common.util import randomString
from plenum.common.constants import VALIDATOR, STEWARD_STRING, POOL_LEDGER_ID, DOMAIN_LEDGER_ID, IDR_CACHE_LABEL, \
    ATTRIB_LABEL, TRUSTEE, STEWARD, KeyValueStorageType, BLS_LABEL, TRUSTEE_STRING
from plenum.test.helper import sdk_get_and_check_replies, waitForViewChange
from plenum.test.test_node import checkNodesConnected, ensureElectionsDone

# noinspection PyUnresolvedReferences
from plenum.test.conftest import tdir as plenum_tdir, nodeReg, \
    whitelist, concerningLogLevels, logcapture, \
    tdirWithPoolTxns, tdirWithDomainTxns, \
    txnPoolNodeSet, \
    poolTxnData, dirName, poolTxnNodeNames, allPluginsPath, tdirWithNodeKeepInited, \
    poolTxnStewardData, poolTxnStewardNames, getValueFromModule, \
    patchPluginManager, txnPoolNodesLooper, warncheck, \
    warnfilters as plenum_warnfilters, do_post_node_creation

# noinspection PyUnresolvedReferences
from plenum.test.conftest import sdk_pool_handle as plenum_pool_handle, sdk_pool_data, sdk_wallet_steward, \
    sdk_wallet_handle, sdk_wallet_data, sdk_steward_seed, sdk_wallet_client, sdk_wallet_trustee, \
    sdk_trustee_seed, trustee_data, sdk_client_seed, poolTxnClientData, poolTxnClientNames, \
    sdk_wallet_stewards, create_node_and_not_start, sdk_wallet_handle, sdk_wallet_new_client, sdk_new_client_seed

from indy_common import strict_types
from indy_common.constants import APP_NAME, CONFIG_LEDGER_ID, CONFIG_LEDGER_AUTH_POLICY, NETWORK_MONITOR, ENDORSER
from indy_common.config_helper import NodeConfigHelper

# noinspection PyUnresolvedReferences
from indy_common.test.conftest import general_conf_tdir, tconf as _tconf, poolTxnTrusteeNames, \
    domainTxnOrderedFields, looper, setTestLogLevel, node_config_helper_class, config_helper_class

from indy_node.test.helper import TestNode, TestNodeBootstrap
from indy_node.test.mock import build_nym_request

from indy_node.server.upgrader import Upgrader
from indy_node.utils.node_control_utils import NodeControlUtil

from indy_node.test.upgrade.helper import releaseVersion

# typecheck during tests
strict_types.defaultShouldCheck = True

Logger.setLogLevel(logging.NOTSET)


@pytest.fixture(scope="module")
def tconf(_tconf):
    oldMax3PCBatchSize = _tconf.Max3PCBatchSize
    _tconf.Max3PCBatchSize = 1
    yield _tconf
    _tconf.Max3PCBatchSize = oldMax3PCBatchSize

@pytest.fixture(scope='module')
def sdk_pool_handle(plenum_pool_handle, nodeSet):
    return plenum_pool_handle


@pytest.fixture(scope="session")
def warnfilters():
    def _():
        warnings.filterwarnings(
            'ignore',
            category=DeprecationWarning,
            module='indy_common\.persistence\.identity_graph',
            message="The 'warn' method is deprecated")
        warnings.filterwarnings(
            'ignore', category=ResourceWarning, message='unclosed transport')

    return _


@pytest.fixture(scope='module', name="sdk_node_theta_added")
def sdk_node_theta_added_fixture(looper,
                         txnPoolNodeSet,
                         tdir,
                         tconf,
                         sdk_pool_handle,
                         sdk_wallet_trustee,
                         allPluginsPath,
                         node_config_helper_class,
                         testNodeClass,
                         name=None,
                         services=[VALIDATOR]):
    return sdk_node_theta_added(looper, txnPoolNodeSet, tdir, tconf, sdk_pool_handle, sdk_wallet_trustee, allPluginsPath, node_config_helper_class, testNodeClass, name)


def sdk_node_theta_added(looper,
                         txnPoolNodeSet,
                         tdir,
                         tconf,
                         sdk_pool_handle,
                         sdk_wallet_trustee,
                         allPluginsPath,
                         node_config_helper_class,
                         testNodeClass,
                         name=None,
                         services=[VALIDATOR]):
    new_steward_name = "testClientSteward" + randomString(3)
    new_node_name = name or "Theta"

    new_steward_wallet = sdk_add_new_nym(looper,
                                         sdk_pool_handle,
                                         sdk_wallet_trustee,
                                         alias=new_steward_name,
                                         role=STEWARD_STRING)

    sigseed, verkey, bls_key, nodeIp, nodePort, clientIp, clientPort, key_proof = \
        prepare_new_node_data(tconf, tdir, new_node_name,
                              configClass=node_config_helper_class)

    # filling node request
    _, steward_did = new_steward_wallet
    node_request = looper.loop.run_until_complete(
        prepare_node_request(steward_did,
                             new_node_name=new_node_name,
                             clientIp=clientIp,
                             clientPort=clientPort,
                             nodeIp=nodeIp,
                             nodePort=nodePort,
                             bls_key=bls_key,
                             sigseed=sigseed,
                             services=services,
                             key_proof=key_proof))

    # sending request using 'sdk_' functions
    request_couple = sdk_sign_and_send_prepared_request(looper, new_steward_wallet,
                                                        sdk_pool_handle, node_request)

    # waitng for replies
    sdk_get_and_check_replies(looper, [request_couple])

    new_node = create_and_start_new_node(looper, new_node_name, tdir, sigseed,
                                         (nodeIp, nodePort), (clientIp, clientPort),
                                         tconf, True, allPluginsPath,
                                         testNodeClass,
                                         configClass=node_config_helper_class)

    if services == [VALIDATOR]:
        txnPoolNodeSet.append(new_node)
        looper.run(checkNodesConnected(txnPoolNodeSet))
    sdk_pool_refresh(looper, sdk_pool_handle)
    return new_steward_wallet, new_node


@pytest.fixture(scope="module")
def sdk_wallet_endorser_factory(looper, sdk_pool_handle, sdk_wallet_trustee):
    def _sdk_wallet_endorser_factory(diddoc_content=None, version=None):
        seed = randomString(32)
        alias = randomString(5)

        raw_nym_request, did = looper.loop.run_until_complete(
            prepare_nym_request(sdk_wallet_trustee, seed, alias, role='ENDORSER')
        )
        nym_request = json.loads(raw_nym_request)
        if diddoc_content:
            nym_request["operation"]["diddocContent"] = (
                json.dumps(diddoc_content)
                if isinstance(diddoc_content, dict)
                else diddoc_content
            )
        if version:
            nym_request["operation"]["version"] = version

        request_couple = sdk_sign_and_send_prepared_request(
            looper, sdk_wallet_trustee, sdk_pool_handle, json.dumps(nym_request)
        )
        sdk_get_and_check_replies(looper, [request_couple])

        return sdk_wallet_trustee[0], did

    return _sdk_wallet_endorser_factory


@pytest.fixture(scope="module")
def sdk_wallet_endorser(looper, sdk_pool_handle, sdk_wallet_trustee):
    return sdk_add_new_nym(looper, sdk_pool_handle, sdk_wallet_trustee,
                           alias='TA-1', role='ENDORSER')


@pytest.fixture(scope="module")
def sdk_user_wallet_a(nodeSet, sdk_wallet_endorser,
                      sdk_pool_handle, looper):
    return sdk_add_new_nym(looper, sdk_pool_handle,
                           sdk_wallet_endorser, alias='userA',
                           skipverkey=True)


# patch that makes sense in general for tests
# since '_get_curr_info' relies on OS package manager
@pytest.fixture(scope="module")
def patchNodeControlUtil():
    old__get_curr_info = getattr(NodeControlUtil, '_get_curr_info')

    @classmethod
    def _get_curr_info(cls, package):
        from stp_core.common.log import getlogger
        import os
        logger = getlogger()
        if package == APP_NAME:
            return (
                "Package: {}\nStatus: install ok installed\nPriority: extra\nSection: default\n"
                "Installed-Size: 21\nMaintainer: maintainer\nArchitecture: amd64\nVersion: {}\n"
            ).format(APP_NAME, releaseVersion())

        raise ValueError("Only {} is expected, got: {}".format(APP_NAME, package))

    setattr(NodeControlUtil, '_get_curr_info', _get_curr_info)
    yield
    setattr(NodeControlUtil, '_get_curr_info', old__get_curr_info)


# link patching with tdir as the most common fixture to make the patch
# applied regardless usage of the pool (there are cases when node control
# is tested without pool creation)
@pytest.fixture(scope="module")
def tdir(patchNodeControlUtil, plenum_tdir):
    return plenum_tdir


@pytest.fixture(scope="module")
def nodeSet(txnPoolNodeSet):
    return txnPoolNodeSet


@pytest.fixture(scope="module")
def testNodeClass():
    return TestNode


@pytest.fixture(scope="module")
def testNodeBootstrapClass():
    return TestNodeBootstrap


@pytest.fixture(scope="module")
def newNodeAdded(looper, nodeSet, tdir, tconf, sdk_pool_handle,
                 sdk_wallet_trustee, allPluginsPath):
    view_no = nodeSet[0].viewNo
    new_steward_wallet, new_node = sdk_node_theta_added(looper,
                                                        nodeSet,
                                                        tdir,
                                                        tconf,
                                                        sdk_pool_handle,
                                                        sdk_wallet_trustee,
                                                        allPluginsPath,
                                                        node_config_helper_class=NodeConfigHelper,
                                                        testNodeClass=TestNode,
                                                        name='')
    waitForViewChange(looper=looper, txnPoolNodeSet=nodeSet,
                      expectedViewNo=view_no + 1)
    ensureElectionsDone(looper=looper, nodes=nodeSet)
    return new_steward_wallet, new_node


@pytest.fixture(scope='module')
def nodeIds(nodeSet):
    return next(iter(nodeSet)).poolManager.nodeIds


@pytest.fixture(scope="module")
def pool_ledger(tconf, tmpdir_factory):
    tdir = tmpdir_factory.mktemp('tmp').strpath
    return Ledger(CompactMerkleTree(),
                  dataDir=tdir)


@pytest.fixture(scope="module")
def domain_ledger(tconf, tmpdir_factory):
    tdir = tmpdir_factory.mktemp('tmp').strpath
    return Ledger(CompactMerkleTree(),
                  dataDir=tdir)


@pytest.fixture(scope="module")
def config_ledger(tconf, tmpdir_factory):
    tdir = tmpdir_factory.mktemp('tmp').strpath
    return Ledger(CompactMerkleTree(),
                  dataDir=tdir)


@pytest.fixture(scope="module")
def config_state():
    return PruningState(KeyValueStorageInMemory())


@pytest.fixture(scope="module")
def domain_state():
    return PruningState(KeyValueStorageInMemory())


@pytest.fixture(scope="module")
def pool_state():
    return PruningState(KeyValueStorageInMemory())


@pytest.fixture(scope="module")
def idr_cache_store(idr_cache):
    return idr_cache


@pytest.fixture(scope="module")
def attrib_store():
    return AttributeStore(KeyValueStorageInMemory())


@pytest.fixture(scope="module")
def bls_store():
    return BlsStore(key_value_type=KeyValueStorageType.Memory,
                    data_location=None,
                    key_value_storage_name="BlsInMemoryStore",
                    serializer=serialization.multi_sig_store_serializer)


@pytest.fixture(scope='module')
def idr_cache():
    cache = IdrCache("Cache",
                     KeyValueStorageInMemory())
    i = 0

    for id in IDENTIFIERS[TRUSTEE]:
        i += 1
        cache.set(id, i, int(time.time()), role=TRUSTEE,
                  verkey="trustee_identifier_verkey", isCommitted=False)

    for id in IDENTIFIERS[STEWARD]:
        i += 1
        cache.set(id, i, int(time.time()), role=STEWARD,
                  verkey="steward_identifier_verkey", isCommitted=False)

    for id in IDENTIFIERS[ENDORSER]:
        i += 1
        cache.set(id, i, int(time.time()), role=ENDORSER,
                  verkey="endorser_identifier_verkey", isCommitted=False)

    for id in IDENTIFIERS[NETWORK_MONITOR]:
        i += 1
        cache.set(id, i, int(time.time()), role=NETWORK_MONITOR,
                  verkey="network_monitor_identifier_verkey", isCommitted=False)

    for id in IDENTIFIERS["OtherRole"]:
        i += 1
        cache.set(id, i, int(time.time()), role='OtherRole',
                  verkey="other_verkey", isCommitted=False)

    for id in IDENTIFIERS[None]:
        i += 1
        cache.set(id, i, int(time.time()), role=None,
                  verkey="identity_owner_verkey", isCommitted=False)

    return cache


@pytest.fixture(scope="module")
def constraint_serializer():
    return ConstraintsSerializer(domain_state_serializer)


@pytest.fixture(scope='module')
def write_auth_req_validator(idr_cache,
                             constraint_serializer,
                             config_state):
    validator = WriteRequestValidator(config=FakeSomething(authPolicy=CONFIG_LEDGER_AUTH_POLICY),
                                      auth_map=auth_map,
                                      cache=idr_cache,
                                      config_state=config_state,
                                      state_serializer=constraint_serializer)
    return validator


@pytest.fixture(scope="module")
def db_manager(pool_state, domain_state, config_state,
               pool_ledger, domain_ledger, config_ledger,
               idr_cache_store,
               attrib_store,
               bls_store):
    dbm = DatabaseManager()
    dbm.register_new_database(POOL_LEDGER_ID, pool_ledger, pool_state)
    dbm.register_new_database(DOMAIN_LEDGER_ID, domain_ledger, domain_state)
    dbm.register_new_database(CONFIG_LEDGER_ID, config_ledger, config_state)
    dbm.register_new_store(IDR_CACHE_LABEL, idr_cache_store)
    dbm.register_new_store(ATTRIB_LABEL, attrib_store)
    dbm.register_new_store(BLS_LABEL, bls_store)
    return dbm


@pytest.fixture(scope="module")
def fake_pool_cfg():
    return FakeSomething()


@pytest.fixture(scope="module")
def fake_upgrader():
    return FakeSomething()


@pytest.fixture(scope="module")
def fake_restarter():
    return FakeSomething()


@pytest.fixture(scope="module")
def fake_pool_manager():
    return FakeSomething()


@pytest.fixture(scope="module")
def fake_node(db_manager, fake_pool_cfg, fake_upgrader, fake_restarter, fake_pool_manager):
    wm = WriteRequestManager(database_manager=db_manager)
    rm = ReadRequestManager()
    am = ActionRequestManager()
    return FakeSomething(name="fake_node",
                         dataLocation="//place_that_cannot_exist",
                         genesis_dir="//place_that_cannot_exist",
                         ledger_ids=Node.ledger_ids,
                         write_manager=wm,
                         read_manager=rm,
                         action_manager=am,
                         db_manager=db_manager,
                         write_req_validator=write_auth_req_validator,
                         config=FakeSomething(
                             stewardThreshold=20,
                             poolTransactionsFile="//pool_genesis_that_cannot_exist",
                             domainTransactionsFile="//domain_genesis_that_cannot_exist"
                         ),
                         poolCfg=fake_pool_cfg,
                         upgrader=fake_upgrader,
                         restarter=fake_restarter,
                         poolManager=fake_pool_manager)


@pytest.fixture(scope="module")
def _managers(write_auth_req_validator,
              fake_node):
    nbs = TestNodeBootstrap(fake_node)
    nbs._register_domain_req_handlers()
    nbs._register_domain_batch_handlers()
    nbs._register_config_req_handlers()
    nbs._register_config_batch_handlers()
    return fake_node.write_manager, fake_node.read_manager


@pytest.fixture(scope="module")
def write_manager(_managers):
    return _managers[0]


@pytest.fixture(scope="module")
def read_manager(_managers):
    return _managers[1]


@pytest.fixture(scope='module')
def write_request_validation(write_auth_req_validator):
    def wrapped(*args, **kwargs):
        try:
            write_auth_req_validator.validate(*args, **kwargs)
        except UnauthorizedClientRequest:
            return False
        return True

    return wrapped


@pytest.fixture(scope='module')
def sdk_wallet_trustee_list(looper,
                            sdk_wallet_trustee,
                            sdk_pool_handle):
    sdk_wallet_trustee_list = []
    for i in range(3):
        wallet = sdk_add_new_nym(looper,
                                 sdk_pool_handle,
                                 sdk_wallet_trustee,
                                 alias='trustee{}'.format(i),
                                 role=TRUSTEE_STRING)
        sdk_wallet_trustee_list.append(wallet)
    return sdk_wallet_trustee_list



import pytest

from indy_common.constants import POOL_UPGRADE, ACTION, START, SCHEDULE, NODE_UPGRADE, COMPLETE
from indy_node.server.txn_version_controller import TxnVersionController
from plenum.common.constants import TXN_PAYLOAD_TYPE, TXN_PAYLOAD, TXN_PAYLOAD_DATA, DATA, VERSION, \
    TXN_PAYLOAD_METADATA_FROM, TXN_PAYLOAD_METADATA, TXN_METADATA, TXN_METADATA_TIME

node_count = 4


@pytest.fixture()
def txn_version_controller():
    return TxnVersionController()


@pytest.fixture()
def pool_upgrade_txn():
    return {TXN_METADATA: {TXN_METADATA_TIME: 0},
            TXN_PAYLOAD: {TXN_PAYLOAD_TYPE: POOL_UPGRADE,
                          TXN_PAYLOAD_DATA: {ACTION: START,
                                             SCHEDULE: {i for i in list(range(node_count))}}}}


def create_node_upgrade_txn(version, frm, timestamp=0):
    return {TXN_METADATA: {TXN_METADATA_TIME: timestamp},
            TXN_PAYLOAD: {TXN_PAYLOAD_TYPE: NODE_UPGRADE,
                          TXN_PAYLOAD_METADATA: {TXN_PAYLOAD_METADATA_FROM: frm},
                          TXN_PAYLOAD_DATA: {DATA: {ACTION: COMPLETE,
                                                    VERSION: version}}}}


def test_update_version_without_pool_upgrade(txn_version_controller):
    version = "version1"
    assert txn_version_controller.version is None

    txn_version_controller.update_version(create_node_upgrade_txn(version, "Node1"))
    assert txn_version_controller.version == version


def test_update_version_with_pool_upgrade(txn_version_controller, pool_upgrade_txn):
    version = "version1"
    assert txn_version_controller.version is None

    txn_version_controller.update_version(pool_upgrade_txn)
    txn_version_controller.update_version(create_node_upgrade_txn(version, "Node1"))
    assert txn_version_controller.version is None

    txn_version_controller.update_version(create_node_upgrade_txn(version, "Node1"))
    assert txn_version_controller.version is None

    txn_version_controller.update_version(create_node_upgrade_txn(version, "Node2"))
    assert txn_version_controller.version == version


def test_cleanup_after_update_version(txn_version_controller, pool_upgrade_txn):
    version1 = "version1"
    version2 = "version2"
    assert txn_version_controller.version is None

    txn_version_controller.update_version(pool_upgrade_txn)
    # send NODE_UPGRADE for the old version, check that the version didn't changed (doesn't has a quorum)
    txn_version_controller.update_version(create_node_upgrade_txn(version1, "Node1"))
    assert txn_version_controller.version is None

    # send NODE_UPGRADE for the new version, check that the version changed
    txn_version_controller.update_version(create_node_upgrade_txn(version2, "Node1"))
    txn_version_controller.update_version(create_node_upgrade_txn(version2, "Node2"))
    assert txn_version_controller.version == version2

    # send NODE_UPGRADE for the old version, check that the version didn't change (doesn't has a quorum)
    txn_version_controller.update_version(create_node_upgrade_txn(version1, "Node2"))
    assert txn_version_controller.version == version2


def test_get_version_with_timestamp(txn_version_controller, pool_upgrade_txn):
    version1 = "version1"
    version2 = "version2"

    txn_version_controller.update_version(create_node_upgrade_txn(version1, "Node1", 10))
    txn_version_controller.update_version(create_node_upgrade_txn(version2, "Node1", 100))

    assert txn_version_controller.get_pool_version(0) is None
    assert txn_version_controller.get_pool_version(11) == version1
    assert txn_version_controller.get_pool_version(200) == version2



from typing import Optional

from plenum.common.exceptions import InvalidClientRequest
from indy_common.constants import NODE_UPGRADE, CONFIG_LEDGER_ID
from indy_common.types import Request
from plenum.server.database_manager import DatabaseManager
from plenum.server.request_handlers.handler_interfaces.write_request_handler import WriteRequestHandler


class NodeUpgradeHandler(WriteRequestHandler):
    def __init__(self, database_manager: DatabaseManager):
        super(NodeUpgradeHandler, self).__init__(database_manager=database_manager,
                                                 txn_type=NODE_UPGRADE,
                                                 ledger_id=CONFIG_LEDGER_ID)

    def update_state(self, txn, prev_result, request, is_committed=False):
        pass

    def additional_dynamic_validation(self, request: Request, req_pp_time: Optional[int]):
        pass

    def static_validation(self, request: Request):
        raise InvalidClientRequest(request.identifier, request.reqId,
                                   "External NODE_UPGRADE requests are not allowed")



from indy_common.constants import START
from indy_node.test.pool_restart.test_pool_restart import pool_restart_now


def test_pool_restart_now_with_empty_datetime(
        sdk_pool_handle, sdk_wallet_trustee, looper, tdir, tconf, txnPoolNodeSet):
    pool_restart_now(sdk_pool_handle, sdk_wallet_trustee, looper,
                     tdir, tconf, START, "", use_time=False)



from plenum.common.constants import STEWARD, TRUSTEE
from indy_common.constants import ENDORSER
from indy_common.roles import Roles


def testRolesAreEncoded():
    assert STEWARD == "2"
    assert TRUSTEE == "0"
    assert ENDORSER == "101"


def testRolesEnumDecoded():
    assert Roles.STEWARD.name == "STEWARD"
    assert Roles.TRUSTEE.name == "TRUSTEE"
    assert Roles.ENDORSER.name == "ENDORSER"


def testRolesEnumEncoded():
    assert Roles.STEWARD.value == "2"
    assert Roles.TRUSTEE.value == "0"
    assert Roles.ENDORSER.value == "101"


def testNameFromValue():
    assert Roles.nameFromValue("2") == "STEWARD"
    assert Roles.nameFromValue("0") == "TRUSTEE"
    assert Roles.nameFromValue("101") == "ENDORSER"
    assert Roles.nameFromValue(None) == "None role"



from storage.kv_store import KeyValueStorage


class AttributeStore:
    """
    Stores attributes as key value pair where the key is hash of the
    attribute as stored in ledger and value is the actual value if the attribute
    """

    def __init__(self, keyValueStorage: KeyValueStorage):
        self._keyValueStorage = keyValueStorage

    def set(self, key, value):
        self._keyValueStorage.put(key, value)

    def get(self, key):
        val = self._keyValueStorage.get(key)
        return val.decode()

    def remove(self, key):
        self._keyValueStorage.remove(key)

    def close(self):
        self._keyValueStorage.close()



import pytest
from indy_common.authorize.auth_actions import EDIT_PREFIX, ADD_PREFIX

from indy_common.constants import OLD_VALUE, AUTH_TYPE, ROLE, ENDORSER, AUTH_ACTION
from indy_node.server.request_handlers.config_req_handlers.auth_rule.abstract_auth_rule_handler import \
    AbstractAuthRuleHandler

from indy_node.server.request_handlers.config_req_handlers.auth_rule.auth_rule_handler import AuthRuleHandler
from indy_node.server.request_handlers.config_req_handlers.auth_rule.static_auth_rule_helper import StaticAuthRuleHelper
from indy_node.test.auth_rule.helper import generate_auth_rule_operation
from indy_node.test.request_handlers.helper import add_to_idr
from plenum.common.constants import STEWARD, TRUSTEE
from plenum.common.exceptions import InvalidClientRequest, UnauthorizedClientRequest
from plenum.common.txn_util import reqToTxn, get_payload_data


@pytest.fixture(scope="module")
def auth_rule_handler(db_manager, write_auth_req_validator):
    return AuthRuleHandler(db_manager, write_auth_req_validator)


def test_auth_rule_static_validation(auth_rule_request, auth_rule_handler: AuthRuleHandler):
    auth_rule_handler.static_validation(auth_rule_request)


def test_auth_rule_static_validation_failed_without_old_value(auth_rule_request,
                                                              auth_rule_handler: AuthRuleHandler):
    if OLD_VALUE in auth_rule_request.operation:
        del auth_rule_request.operation[OLD_VALUE]
    auth_rule_request.operation[AUTH_ACTION] = EDIT_PREFIX
    with pytest.raises(InvalidClientRequest, match="EDIT must contain field old_value"):
        auth_rule_handler.static_validation(auth_rule_request)


def test_auth_rule_static_validation_failed_with_excess_field(auth_rule_request,
                                                              auth_rule_handler: AuthRuleHandler):
    auth_rule_request.operation[OLD_VALUE] = "old_value"
    auth_rule_request.operation[AUTH_TYPE] = ADD_PREFIX
    with pytest.raises(InvalidClientRequest, match="ADD must not contain field old_value"):
        auth_rule_handler.static_validation(auth_rule_request)


def test_auth_rule_static_validation_failed_with_incorrect_key(auth_rule_request,
                                                               auth_rule_handler: AuthRuleHandler):
    auth_rule_request.operation = generate_auth_rule_operation(auth_action=ADD_PREFIX,
                                                               auth_type="wrong_type",
                                                               field=ROLE, new_value=ENDORSER)
    with pytest.raises(InvalidClientRequest, match="key .* is not found in authorization map"):
        auth_rule_handler.static_validation(auth_rule_request)


def test_auth_rule_dynamic_validation_without_permission(auth_rule_request,
                                                         auth_rule_handler: AuthRuleHandler, creator):
    add_to_idr(auth_rule_handler.database_manager.idr_cache, creator, STEWARD)
    with pytest.raises(UnauthorizedClientRequest):
        auth_rule_handler.dynamic_validation(auth_rule_request, 0)


def test_auth_rule_dynamic_validation(auth_rule_request,
                                      auth_rule_handler: AuthRuleHandler, creator):
    add_to_idr(auth_rule_handler.database_manager.idr_cache, creator, TRUSTEE)
    auth_rule_handler.dynamic_validation(auth_rule_request, 0)


def test_update_state(auth_rule_request, auth_rule_handler: AuthRuleHandler):
    txn = reqToTxn(auth_rule_request)
    payload = get_payload_data(txn)
    constraint = StaticAuthRuleHelper.get_auth_constraint(payload)
    auth_key = StaticAuthRuleHelper.get_auth_key(payload)
    path = AbstractAuthRuleHandler.make_state_path_for_auth_rule(auth_key)

    auth_rule_handler.update_state(txn, None, auth_rule_request)
    assert auth_rule_handler.get_from_state(path) == constraint.as_dict




#!/usr/bin/python3

import importlib
import os
import pkgutil
import distro
import subprocess
import time
from pathlib import Path
from functools import cmp_to_key

from stp_core.common.log import getlogger

from indy_common.util import compose_cmd
from indy_node.server.upgrader import Upgrader

SCRIPT_PREFIX = Path('data', 'migrations')
PLATFORM_PREFIX = {
    'Ubuntu': 'deb'
}

logger = getlogger()


def _call_migration_script(migration_script, current_platform, timeout):
    migration_script_path = \
        os.path.normpath(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                '..',
                '..',
                str(SCRIPT_PREFIX),
                PLATFORM_PREFIX[current_platform],
                migration_script + '.py'))
    logger.info('script path {}'.format(migration_script_path))
    ret = subprocess.run(
        compose_cmd(
            ['python3 {}'.format(migration_script_path)]
        ),
        shell=True,
        timeout=timeout)
    if ret.returncode != 0:
        msg = 'Migration failed: script returned {}'.format(ret.returncode)
        logger.error(msg)
        raise Exception(msg)


# Returns number of performed migrations
def migrate(current_version, new_version, timeout):
    current_platform = _get_current_platform()
    logger.info('Migrating from {} to {} on {}'.format(
        current_version, new_version, current_platform))
    migration_scripts = _get_migration_scripts(current_platform)
    logger.debug('Found migration scripts: {}'.format(migration_scripts))
    migration_scripts = _get_relevant_migrations(
        migration_scripts, current_version, new_version)
    if not len(migration_scripts):
        logger.info('No migrations can be applied to the current code.')
        return 0
    logger.info('Following migrations will be applied: {}'.format(
        migration_scripts))
    for migration in migration_scripts:
        logger.info('Applying migration {}'.format(migration))
        start_time = time.time()
        _call_migration_script(migration, current_platform, timeout)
        logger.info('Migration {} applied in {} seconds'.format(
            migration, time.time() - start_time))
    return len(migration_scripts)


def _get_migration_scripts(current_platform):
    # Data folder is published as a separate 'data' python package
    migrations = importlib.import_module(
        '.'.join(SCRIPT_PREFIX.parts + (PLATFORM_PREFIX[current_platform],))
    )
    return [name
            for module_finder, name, ispkg
            in pkgutil.iter_modules(migrations.__path__)]


def _get_relevant_migrations(migration_scripts, current_version, new_version):
    relevant_migrations = []
    for migration in migration_scripts:
        migration_original_version, migration_new_version = _get_migration_versions(
            migration)
        if not migration_original_version or not migration_new_version:
            continue

        if Upgrader.compareVersions(new_version, current_version) >= 0:
            if Upgrader.compareVersions(
                    migration_original_version, migration_new_version) > 0:
                continue
            if Upgrader.compareVersions(current_version, migration_original_version) <= 0 \
                    and Upgrader.compareVersions(new_version, migration_new_version) >= 0:
                relevant_migrations.append(migration)
        else:
            if Upgrader.compareVersions(
                    migration_new_version, migration_original_version) > 0:
                continue
            if Upgrader.compareVersions(current_version, migration_original_version) >= 0 \
                    and Upgrader.compareVersions(new_version, migration_new_version) <= 0:
                relevant_migrations.append(migration)
        relevant_migrations = sorted(
            relevant_migrations, key=cmp_to_key(_compare_migration_scripts))
    return relevant_migrations


def _get_migration_versions(migration):
    migration_split = migration.split('_')
    if len(migration_split) != 7:
        return None, None
    migration_original_version = '.'.join(migration_split[0:3])
    migration_new_version = '.'.join(migration_split[-3:])
    return migration_original_version, migration_new_version


def _compare_migration_scripts(migration1, migration2):
    migration_original_version1, migration_new_version1 = _get_migration_versions(
        migration1)
    migration_original_version2, migration_new_version2 = _get_migration_versions(
        migration2)
    if Upgrader.compareVersions(
            migration_original_version1, migration_original_version2) == 0:
        return Upgrader.compareVersions(
            migration_new_version1, migration_new_version2)
    return Upgrader.compareVersions(
        migration_original_version1, migration_original_version2)


def _get_current_platform():
    name = distro.linux_distribution()[0]
    if 'Ubuntu' in name or 'ubuntu' in name:
        return 'Ubuntu'
    raise Exception('Platform is not supported')



from abc import abstractmethod, ABCMeta

from copy import deepcopy

from indy_common.state import domain
from indy_common.types import Request
from indy_common.constants import REVOC_REG_DEF_ID, VALUE, ACCUM, PREV_ACCUM, ISSUED, REVOKED
from plenum.common.exceptions import InvalidClientRequest
from plenum.common.txn_util import get_from, get_req_id, get_payload_data


class RevocationStrategy(metaclass=ABCMeta):

    def __init__(self, state):
        self.state = state
        self.author_did = None
        self.revoc_reg_def_id = None
        self.req_id = None

    def set_parameters_from_txn(self, author_did, revoc_reg_def_id, req_id):
        self.author_did = author_did
        self.revoc_reg_def_id = revoc_reg_def_id
        self.req_id = req_id

    def validate(self, current_entry, req: Request):
        self.set_parameters_from_txn(author_did=req.identifier,
                                     revoc_reg_def_id=req.operation[REVOC_REG_DEF_ID],
                                     req_id=req.reqId)
        # General checks for all Revocation entries
        operation = req.operation
        value_from_txn = operation.get(VALUE)
        issued_from_txn = value_from_txn.get(ISSUED, [])
        revoked_from_txn = value_from_txn.get(REVOKED, [])
        intersection = set(issued_from_txn).intersection(set(revoked_from_txn))
        if len(intersection) > 0:
            raise InvalidClientRequest(self.author_did,
                                       self.req_id,
                                       "Can not have an index in both "
                                       "'issued' and 'revoked' lists")
        prev_accum = value_from_txn.get(PREV_ACCUM, None)
        accum = value_from_txn.get(ACCUM, None)
        if prev_accum is not None and current_entry is None:
            raise InvalidClientRequest(self.author_did,
                                       self.req_id,
                                       "Got '{}': {} but there is no "
                                       "previous REVOC_REG_ENTRY".format(PREV_ACCUM,
                                                                         prev_accum))
        if current_entry is None:
            return None
        value_from_state = current_entry.get(VALUE)
        assert value_from_state
        current_accum = value_from_state.get(ACCUM)
        if current_accum != value_from_txn.get(PREV_ACCUM):
            raise InvalidClientRequest(self.author_did,
                                       self.req_id,
                                       "The current accumulator value: {} "
                                       "must be equal to the last accumulator "
                                       "value: {} in transaction".format(
                                           current_accum,
                                           value_from_state.get(PREV_ACCUM)))
        if len(issued_from_txn) == 0 and len(revoked_from_txn) == 0:
            raise InvalidClientRequest(self.author_did,
                                       self.req_id,
                                       "Got '{}' and '{}' but '{}' and '{}' lists are empty".format(
                                           PREV_ACCUM, ACCUM,
                                           ISSUED, REVOKED))
        if prev_accum == accum and (len(issued_from_txn) > 0 or len(revoked_from_txn) > 0):
            raise InvalidClientRequest(self.author_did,
                                       self.req_id,
                                       "Got equal accum and prev_accum but "
                                       "issued and revoked indices are not empty")

        # Do strategy specific validation
        self.specific_validation(current_entry, req)

    def set_to_state(self, txn):
        # Set REVOC_REG_ENTRY
        path, value_bytes = domain.prepare_revoc_reg_entry_for_state(txn)
        self.state.set(path, value_bytes)
        # Set ACCUM from REVOC_REG_ENTRY
        txn_data = get_payload_data(txn)
        txn_data[VALUE] = {ACCUM: txn_data[VALUE][ACCUM]}
        path, value_bytes = domain.prepare_revoc_reg_entry_accum_for_state(txn)
        self.state.set(path, value_bytes)

    @abstractmethod
    def specific_validation(self, current_entry, req: Request):
        raise NotImplementedError

    @abstractmethod
    def write(self, current_reg_entry, txn):
        raise NotImplementedError

    @staticmethod
    def get_delta(to_dict, from_dict=None):
        raise NotImplementedError


class RevokedStrategy(RevocationStrategy):
    # This strategy save in state only revoked indices

    def specific_validation(self, current_reg_entry, req: Request):
        value_from_state = current_reg_entry.get(VALUE)
        assert value_from_state
        indices = value_from_state.get(REVOKED, [])
        value_from_txn = req.operation.get(VALUE)
        issued_from_txn = value_from_txn.get(ISSUED, [])
        revoked_from_txn = value_from_txn.get(REVOKED, [])
        issued_difference = set(issued_from_txn).difference(indices)
        if len(issued_difference) > 0:
            raise InvalidClientRequest(self.author_did,
                                       self.req_id,
                                       "Issued indices from txn: {} "
                                       "are not present in the current "
                                       "revoked list from state: {}".format(issued_difference,
                                                                            indices))
        revoked_intersection = set(indices).intersection(revoked_from_txn)
        if len(revoked_intersection) > 0:
            raise InvalidClientRequest(self.author_did,
                                       self.req_id,
                                       "Revoked indices from txn: {} "
                                       "are already revoked "
                                       "in current state: {}".format(revoked_intersection,
                                                                     indices))

    def write(self, current_reg_entry, txn):
        txn = deepcopy(txn)
        txn_data = get_payload_data(txn)
        self.set_parameters_from_txn(author_did=get_from(txn),
                                     revoc_reg_def_id=txn_data.get(REVOC_REG_DEF_ID),
                                     req_id=get_req_id(txn))
        if current_reg_entry is not None:
            value_from_state = current_reg_entry.get(VALUE)
            assert value_from_state
            indices = value_from_state.get(REVOKED, [])
            value_from_txn = txn_data.get(VALUE)
            issued_from_txn = value_from_txn.get(ISSUED, [])
            revoked_from_txn = value_from_txn.get(REVOKED, [])
            # set with all previous revoked minus issued from txn
            result_indicies = set(indices).difference(issued_from_txn)
            result_indicies.update(revoked_from_txn)
            value_from_txn[ISSUED] = []
            value_from_txn[REVOKED] = list(result_indicies)
            txn_data[VALUE] = value_from_txn
        # contains already changed txn
        self.set_to_state(txn)

    @staticmethod
    def get_delta(to_dict, from_dict=None):
        if from_dict is None:
            return to_dict[ISSUED], to_dict[REVOKED]
        revoked = set(to_dict[REVOKED]).difference(set(from_dict[REVOKED]))
        issued = set(from_dict[REVOKED]).difference(set(to_dict[REVOKED]))
        return issued, revoked


class IssuedStrategy(RevocationStrategy):
    # This strategy saves in state only issued indices

    def specific_validation(self, current_reg_entry, req: Request):
        value_from_state = current_reg_entry.get(VALUE)
        assert value_from_state
        indices = value_from_state.get(ISSUED, [])
        value_from_txn = req.operation.get(VALUE)
        issued_from_txn = value_from_txn.get(ISSUED, [])
        revoked_from_txn = value_from_txn.get(REVOKED, [])
        revoked_difference = set(revoked_from_txn).difference(indices)
        if len(revoked_difference) > 0:
            raise InvalidClientRequest(self.author_did,
                                       self.req_id,
                                       "Revoked indices from txn: {} "
                                       "are not present in the current "
                                       "issued list from state: {}".format(revoked_difference,
                                                                           indices))
        issued_intersection = set(indices).intersection(issued_from_txn)
        if len(issued_intersection) > 0:
            raise InvalidClientRequest(self.author_did,
                                       self.req_id,
                                       "Issued indices from txn: {} "
                                       "are already issued "
                                       "in current state: {}".format(issued_intersection,
                                                                     indices))

    def write(self, current_reg_entry, txn):
        txn = deepcopy(txn)
        txn_data = get_payload_data(txn)
        self.set_parameters_from_txn(author_did=get_from(txn),
                                     revoc_reg_def_id=txn_data.get(REVOC_REG_DEF_ID),
                                     req_id=get_req_id(txn))
        if current_reg_entry is not None:
            value_from_state = current_reg_entry.get(VALUE)
            assert value_from_state
            indices = value_from_state.get(ISSUED, [])
            value_from_txn = txn_data.get(VALUE)
            issued_from_txn = value_from_txn.get(ISSUED, [])
            revoked_from_txn = value_from_txn.get(REVOKED, [])
            # set with all previous issued minus revoked from txn
            result_indicies = set(indices).difference(revoked_from_txn)
            result_indicies.update(issued_from_txn)
            value_from_txn[REVOKED] = []
            value_from_txn[ISSUED] = list(result_indicies)
            txn_data[VALUE] = value_from_txn
        # contains already changed txn
        self.set_to_state(txn)

    @staticmethod
    def get_delta(to_dict, from_dict=None):
        if from_dict is None:
            return to_dict[ISSUED], to_dict[REVOKED]
        issued = set(to_dict[ISSUED]).difference(set(from_dict[ISSUED]))
        revoked = set(from_dict[ISSUED]).difference(set(to_dict[ISSUED]))
        return issued, revoked



from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER, AuthConstraintForbidden
from indy_common.constants import ENDORSER
from indy_common.test.auth.metadata.helper import validate, PLUGIN_FIELD, Action
from plenum.common.constants import TRUSTEE

MAX_SIG_COUNT = 3


def test_plugin_simple_rule_1_sig_owner_no_endorser(write_auth_req_validator, write_request_validation,
                                                    signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraint(role=IDENTITY_OWNER, sig_count=1, need_to_be_owner=True,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=IDENTITY_OWNER, endorser=None, sigs={IDENTITY_OWNER: s},
                   is_owner=True, amount=2, extra_sigs=False)
            for s in range(1, MAX_SIG_COUNT + 1)
        ],
        author=IDENTITY_OWNER, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_1_sig_owner_endorser(write_auth_req_validator, write_request_validation,
                                                 signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraint(role=IDENTITY_OWNER, sig_count=1, need_to_be_owner=True,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=IDENTITY_OWNER, endorser=ENDORSER, sigs={IDENTITY_OWNER: s1, ENDORSER: s2},
                   is_owner=True, amount=2, extra_sigs=True)
            for s1 in range(1, MAX_SIG_COUNT + 1)
            for s2 in range(1, MAX_SIG_COUNT + 1)
        ],
        author=IDENTITY_OWNER, endorser=ENDORSER,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_1_sig_endorser_no_endorser(write_auth_req_validator, write_request_validation,
                                                       signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraint(role=ENDORSER, sig_count=1, need_to_be_owner=True,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=ENDORSER, endorser=None, sigs={ENDORSER: s},
                   is_owner=True, amount=2, extra_sigs=True)
            for s in range(1, MAX_SIG_COUNT + 1)
        ],
        author=ENDORSER, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_1_sig_endorser_endorser(write_auth_req_validator, write_request_validation,
                                                    signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraint(role=ENDORSER, sig_count=1, need_to_be_owner=True,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=ENDORSER, endorser=ENDORSER, sigs={ENDORSER: s},
                   is_owner=True, amount=2, extra_sigs=True)
            for s in range(1, MAX_SIG_COUNT + 1)
        ],
        author=ENDORSER, endorser=ENDORSER,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_1_sig_trustee_no_endorser(write_auth_req_validator, write_request_validation,
                                                      signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraint(role=ENDORSER, sig_count=1, need_to_be_owner=True,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=TRUSTEE, endorser=None, sigs={ENDORSER: s1, TRUSTEE: s2},
                   is_owner=True, amount=2, extra_sigs=True)
            for s1 in range(1, MAX_SIG_COUNT + 1)
            for s2 in range(1, MAX_SIG_COUNT + 1)
        ],
        author=TRUSTEE, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_1_sig_trustee_endorser(write_auth_req_validator, write_request_validation,
                                                   signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraint(role=ENDORSER, sig_count=1, need_to_be_owner=True,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=TRUSTEE, endorser=ENDORSER, sigs={TRUSTEE: s2, ENDORSER: s3},
                   is_owner=True, amount=2, extra_sigs=True)
            for s1 in range(1, MAX_SIG_COUNT + 1)
            for s2 in range(1, MAX_SIG_COUNT + 1)
            for s3 in range(1, MAX_SIG_COUNT + 1)
        ],
        author=TRUSTEE, endorser=ENDORSER,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_1_sig_all_roles_owner_no_endorser(write_auth_req_validator, write_request_validation,
                                                              signatures, is_owner, off_ledger_signature, amount):
    validate(
        auth_constraint=AuthConstraint(role='*', sig_count=1, need_to_be_owner=True,
                                       off_ledger_signature=off_ledger_signature,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=IDENTITY_OWNER, endorser=None, sigs={IDENTITY_OWNER: s},
                   is_owner=True, amount=2, extra_sigs=False)
            for s in range(1, MAX_SIG_COUNT + 1)
        ],
        author=IDENTITY_OWNER, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_1_sig_all_roles_owner_endorser(write_auth_req_validator, write_request_validation,
                                                           signatures, is_owner, off_ledger_signature, amount):
    validate(
        auth_constraint=AuthConstraint(role='*', sig_count=1, need_to_be_owner=True,
                                       off_ledger_signature=off_ledger_signature,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=IDENTITY_OWNER, endorser=ENDORSER, sigs={IDENTITY_OWNER: s1, ENDORSER: s2},
                   is_owner=True, amount=2, extra_sigs=True)
            for s1 in range(1, MAX_SIG_COUNT + 1)
            for s2 in range(1, MAX_SIG_COUNT + 1)
        ],
        author=IDENTITY_OWNER, endorser=ENDORSER,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_1_sig_all_roles_trustee_no_endorser(write_auth_req_validator, write_request_validation,
                                                                signatures, is_owner, off_ledger_signature, amount):
    validate(
        auth_constraint=AuthConstraint(role='*', sig_count=1, need_to_be_owner=True,
                                       off_ledger_signature=off_ledger_signature,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: s1},
                   is_owner=True, amount=2, extra_sigs=True)
            for s1 in range(1, MAX_SIG_COUNT + 1)
        ],
        author=TRUSTEE, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_1_sig_all_roles_trustee_endorser(write_auth_req_validator, write_request_validation,
                                                             signatures, is_owner, off_ledger_signature, amount):
    validate(
        auth_constraint=AuthConstraint(role='*', sig_count=1, need_to_be_owner=True,
                                       off_ledger_signature=off_ledger_signature,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=TRUSTEE, endorser=ENDORSER, sigs={TRUSTEE: s1, ENDORSER: s2},
                   is_owner=True, amount=2, extra_sigs=True)
            for s1 in range(1, MAX_SIG_COUNT + 1)
            for s2 in range(1, MAX_SIG_COUNT + 1)
        ],
        author=TRUSTEE, endorser=ENDORSER,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_3_sig_trustee_no_endorser(write_auth_req_validator, write_request_validation,
                                                      signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraint(role=TRUSTEE, sig_count=3, need_to_be_owner=False,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=TRUSTEE, endorser=None, sigs={TRUSTEE: 3},
                   is_owner=owner, amount=2, extra_sigs=True)
            for owner in [True, False]],
        author=TRUSTEE, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_3_sig_trustee_endorser(write_auth_req_validator, write_request_validation,
                                                   signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraint(role=TRUSTEE, sig_count=3, need_to_be_owner=False,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=TRUSTEE, endorser=ENDORSER, sigs={TRUSTEE: 3, ENDORSER: s1},
                   is_owner=owner, amount=2, extra_sigs=True)
            for s1 in range(1, MAX_SIG_COUNT + 1)
            for owner in [True, False]
        ],
        author=TRUSTEE, endorser=ENDORSER,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_3_sig_owner_no_endorser(write_auth_req_validator, write_request_validation,
                                                    signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraint(role=TRUSTEE, sig_count=3, need_to_be_owner=False,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[],
        author=IDENTITY_OWNER, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_3_sig_owner_endorser(write_auth_req_validator, write_request_validation,
                                                 signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraint(role=TRUSTEE, sig_count=3, need_to_be_owner=False,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[
            Action(author=IDENTITY_OWNER, endorser=ENDORSER, sigs={TRUSTEE: 3, IDENTITY_OWNER: s1, ENDORSER: s2},
                   is_owner=owner, amount=2, extra_sigs=True)
            for s1 in range(1, MAX_SIG_COUNT + 1)
            for s2 in range(1, MAX_SIG_COUNT + 1)
            for owner in [True, False]
        ],
        author=IDENTITY_OWNER, endorser=ENDORSER,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_0_sig_owner_no_endorser(write_auth_req_validator, write_request_validation,
                                                    signatures, is_owner, off_ledger_signature, amount):
    validate(
        auth_constraint=AuthConstraint(role='*', sig_count=0, need_to_be_owner=False,
                                       off_ledger_signature=off_ledger_signature,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[Action(author=IDENTITY_OWNER, endorser=None, sigs={},
                              is_owner=owner, amount=2, extra_sigs=False)
                       for owner in [True, False]] + [
            Action(author=IDENTITY_OWNER, endorser=None, sigs={IDENTITY_OWNER: s},
                   is_owner=owner, amount=2, extra_sigs=False)
            for owner in [True, False]
            for s in range(1, MAX_SIG_COUNT + 1)],
        author=IDENTITY_OWNER, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_0_sig_owner_endorser(write_auth_req_validator, write_request_validation,
                                                 signatures, is_owner, off_ledger_signature, amount):
    validate(
        auth_constraint=AuthConstraint(role='*', sig_count=0, need_to_be_owner=False,
                                       off_ledger_signature=off_ledger_signature,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[Action(author=IDENTITY_OWNER, endorser=ENDORSER, sigs={ENDORSER: s},
                              is_owner=owner, amount=2, extra_sigs=True)
                       for s in range(1, MAX_SIG_COUNT + 1)
                       for owner in [True, False]],
        author=IDENTITY_OWNER, endorser=ENDORSER,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_0_sig_trustee_no_endorser(write_auth_req_validator, write_request_validation,
                                                      signatures, is_owner, off_ledger_signature, amount):
    validate(
        auth_constraint=AuthConstraint(role='*', sig_count=0, need_to_be_owner=False,
                                       off_ledger_signature=off_ledger_signature,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[Action(author=TRUSTEE, endorser=None, sigs=signature,
                              is_owner=owner, amount=2, extra_sigs=True)
                       for signature in signatures
                       for owner in [True, False]],
        author=TRUSTEE, endorser=None,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_0_sig_trustee_endorser(write_auth_req_validator, write_request_validation,
                                                   signatures, is_owner, off_ledger_signature, amount):
    validate(
        auth_constraint=AuthConstraint(role='*', sig_count=0, need_to_be_owner=False,
                                       off_ledger_signature=off_ledger_signature,
                                       metadata={PLUGIN_FIELD: 2}),
        valid_actions=[Action(author=TRUSTEE, endorser=ENDORSER, sigs={ENDORSER: s},
                              is_owner=owner, amount=2, extra_sigs=True)
                       for s in range(1, MAX_SIG_COUNT + 1)
                       for owner in [True, False]],
        author=TRUSTEE, endorser=ENDORSER,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )


def test_plugin_simple_rule_not_allowed(write_auth_req_validator, write_request_validation,
                                        author, endorser, signatures, is_owner, amount):
    validate(
        auth_constraint=AuthConstraintForbidden(),
        valid_actions=[],
        author=author, endorser=endorser,
        all_signatures=signatures, is_owner=is_owner, amount=amount,
        write_auth_req_validator=write_auth_req_validator,
        write_request_validation=write_request_validation
    )



from collections import OrderedDict

from indy_common.types import ClientGetAuthRuleOperation, AuthRuleValueField
from plenum.common.messages.fields import ConstantField, LimitedLengthStringField, ChooseField

EXPECTED_ORDERED_FIELDS = OrderedDict([
    ("type", ConstantField),
    ("auth_action", ChooseField),
    ("auth_type", LimitedLengthStringField),
    ("field", LimitedLengthStringField),
    ("old_value", AuthRuleValueField),
    ("new_value", AuthRuleValueField),
])


def test_has_expected_fields():
    actual_field_names = OrderedDict(ClientGetAuthRuleOperation.schema).keys()
    assert actual_field_names == EXPECTED_ORDERED_FIELDS.keys()


def test_has_expected_validators():
    schema = dict(ClientGetAuthRuleOperation.schema)
    for field, validator in EXPECTED_ORDERED_FIELDS.items():
        assert isinstance(schema[field], validator)


def test_attrib_with_empty_enc_fails():
    validator = AuthRuleValueField()
    value = ''
    validator.validate(value)
    value = None
    validator.validate(value)



from indy_node.test import waits
from indy_node.test.upgrade.helper import checkUpgradeScheduled
from plenum.common.constants import VERSION
from plenum.test.test_node import checkNodesConnected, ensureElectionsDone
from stp_core.loop.eventually import eventually
from indy_common.config_helper import NodeConfigHelper


def testNodeSchedulesUpgradeAfterRestart(upgradeScheduled, looper, nodeSet,
                                         validUpgrade, testNodeClass,
                                         tdir, tconf, allPluginsPath):
    names = []
    while nodeSet:
        node = nodeSet.pop()
        names.append(node.name)
        node.cleanupOnStopping = False
        looper.removeProdable(node)
        node.stop()
        del node

    for nm in names:
        config_helper = NodeConfigHelper(nm, tconf, chroot=tdir)
        node = testNodeClass(nm, config_helper=config_helper,
                             config=tconf, pluginPaths=allPluginsPath)
        looper.add(node)
        nodeSet.append(node)

    looper.run(checkNodesConnected(nodeSet))
    ensureElectionsDone(looper=looper, nodes=nodeSet, retryWait=1)
    looper.run(
        eventually(
            checkUpgradeScheduled,
            nodeSet,
            validUpgrade[VERSION],
            retryWait=1,
            timeout=waits.expectedUpgradeScheduled()))



import pytest
from copy import deepcopy

from indy_common.authorize.auth_actions import ADD_PREFIX, AuthActionEdit, EDIT_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint, IDENTITY_OWNER
from indy_common.constants import POOL_UPGRADE, ACTION, CANCEL, JUSTIFICATION, SCHEDULE, CONFIG_LEDGER_ID
from indy_node.test.auth_rule.auth_framework.basic import AuthTest
from indy_node.test.upgrade.helper import sdk_ensure_upgrade_sent
from plenum.common.exceptions import RequestRejectedException
from plenum.test.pool_transactions.helper import sdk_add_new_nym
from indy_common.authorize import auth_map

from indy_node.test.helper import build_auth_rule_request_json


class StartUpgradeTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)
        self.valid_upgrade = env.valid_upgrade

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()
        for n in self.env.txnPoolNodeSet:
            n.upgrader.handleUpgradeTxn = lambda *args, **kwargs: True

    def run(self):

        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 2. Check, that new auth rule is used
        self.valid_upgrade['name'] += '1'
        sdk_ensure_upgrade_sent(self.looper, self.sdk_pool_handle, self.new_default_wallet, self.valid_upgrade)

        # Step 3. Check that we cannot do txn the old way
        with pytest.raises(RequestRejectedException):
            sdk_ensure_upgrade_sent(self.looper, self.sdk_pool_handle, self.trustee_wallet, self.valid_upgrade)

        # Step 4. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 5. Check, that default auth rule works
        self.valid_upgrade['name'] += '2'
        sdk_ensure_upgrade_sent(self.looper, self.sdk_pool_handle, self.trustee_wallet, self.valid_upgrade)

    def result(self):
        pass

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet, role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=ADD_PREFIX,
            auth_type=POOL_UPGRADE,
            field='action',
            new_value='start',
            constraint=constraint.as_dict
        )


class CancelUpgradeTest(AuthTest):
    def __init__(self, env, action_id):
        super().__init__(env, action_id)
        self.valid_upgrade = env.valid_upgrade

    def prepare(self):
        self.default_auth_rule = self.get_default_auth_rule()
        self.changed_auth_rule = self.get_changed_auth_rule()
        for n in self.env.txnPoolNodeSet:
            n.upgrader.handleUpgradeTxn = lambda *args, **kwargs: True

    def run(self):

        # Step 1. Change auth rule
        self.send_and_check(self.changed_auth_rule, wallet=self.trustee_wallet)

        # Step 2. Check that we cannot do txn the old way
        self.valid_upgrade['name'] += '1'
        sdk_ensure_upgrade_sent(self.looper, self.sdk_pool_handle, self.trustee_wallet, self.valid_upgrade)

        with pytest.raises(RequestRejectedException):
            self.cancel_upgrade(self.valid_upgrade, self.trustee_wallet)

        # Step 3. Check, that new auth rule is used
        self.cancel_upgrade(self.valid_upgrade, self.new_default_wallet)

        # Step 5. Return default auth rule
        self.send_and_check(self.default_auth_rule, wallet=self.trustee_wallet)

        # Step 6. Check, that default auth rule works
        self.valid_upgrade['name'] += '2'
        sdk_ensure_upgrade_sent(self.looper, self.sdk_pool_handle, self.trustee_wallet, self.valid_upgrade)
        self.cancel_upgrade(self.valid_upgrade, self.trustee_wallet)

    def result(self):
        pass

    def get_changed_auth_rule(self):
        self.new_default_wallet = sdk_add_new_nym(self.looper, self.sdk_pool_handle, self.trustee_wallet, role=IDENTITY_OWNER)
        constraint = AuthConstraint(role=IDENTITY_OWNER,
                                    sig_count=1,
                                    need_to_be_owner=False)
        return build_auth_rule_request_json(
            self.looper, self.trustee_wallet[1],
            auth_action=EDIT_PREFIX,
            auth_type=POOL_UPGRADE,
            field='action',
            old_value='start',
            new_value='cancel',
            constraint=constraint.as_dict
        )

    def cancel_upgrade(self, upgrade, sdk_wallet):
        valid_upgrade_copy = deepcopy(upgrade)
        valid_upgrade_copy[ACTION] = CANCEL
        valid_upgrade_copy[JUSTIFICATION] = '"never gonna give you up"'

        valid_upgrade_copy.pop(SCHEDULE, None)
        sdk_ensure_upgrade_sent(self.looper, self.sdk_pool_handle, sdk_wallet, valid_upgrade_copy)



import pytest
import shutil

from common.version import DigitDotVersion

from indy_common.constants import APP_NAME
from indy_common.version import src_version_cls
from indy_node.utils.node_control_utils import (
    NodeControlUtil, ShellError, DebianVersion
)

# TODO
# - conditionally skip all tests for non-debian systems
# - teste _parse_version_deps_from_pkg_mgr_output deeply

generated_commands = []


@pytest.fixture
def catch_generated_commands(monkeypatch):
    generated_commands[:] = []

    def _f(command, *args, **kwargs):
        generated_commands.append(command)
        return ''

    monkeypatch.setattr(NodeControlUtil, 'run_shell_script', _f)
    monkeypatch.setattr(NodeControlUtil, 'run_shell_script_extended', _f)
    monkeypatch.setattr(NodeControlUtil, 'run_shell_command', _f)


def test_generated_cmd_get_curr_info(catch_generated_commands):
    pkg_name = 'some_package'
    # TODO not an API for now
    NodeControlUtil._get_curr_info(pkg_name)
    assert len(generated_commands) == 1
    assert generated_commands[0] == "dpkg -s {}".format(pkg_name)


def test_generated_cmd_get_latest_pkg_version(catch_generated_commands):
    pkg_name = 'some_package'
    NodeControlUtil.get_latest_pkg_version(pkg_name)
    assert len(generated_commands) == 2
    assert generated_commands[0] == "apt update"
    assert generated_commands[1] == (
        "apt-cache show {} | grep -E '^Version: '"
        .format(pkg_name)
    )

    generated_commands[:] = []
    upstream = src_version_cls(pkg_name)('1.2.3')
    NodeControlUtil.get_latest_pkg_version(
        pkg_name, upstream=upstream, update_cache=False)
    assert len(generated_commands) == 1
    assert generated_commands[0] == (
        "apt-cache show {} | grep -E '^Version: '"
        .format(pkg_name)
    )


def test_generated_cmd_get_info_from_package_manager(catch_generated_commands):
    packages = ['package1', 'package2']
    # TODO not an API for now
    NodeControlUtil._get_info_from_package_manager(*packages)
    assert len(generated_commands) == 1
    assert generated_commands[0] == "apt-cache show {}".format(" ".join(packages))


def test_generated_cmd_update_package_cache(catch_generated_commands):
    NodeControlUtil.update_package_cache()
    assert len(generated_commands) == 1
    assert generated_commands[0] == "apt update"


def test_generated_cmd_get_sys_holds(monkeypatch, catch_generated_commands):
    monkeypatch.setattr(shutil, 'which', lambda *_: 'path')
    NodeControlUtil.get_sys_holds()
    assert len(generated_commands) == 1
    assert generated_commands[0] == "apt-mark showhold"


def test_generated_cmd_hold_packages(monkeypatch, catch_generated_commands):
    packages = ['package1', 'package2']
    monkeypatch.setattr(shutil, 'which', lambda *_: 'path')
    NodeControlUtil.hold_packages(packages)
    assert len(generated_commands) == 1
    assert generated_commands[0] == "apt-mark hold {}".format(' '.join(packages))


def test_get_latest_pkg_version_invalid_args():
    pkg_name = 'any_package'
    with pytest.raises(TypeError) as excinfo:
        NodeControlUtil.get_latest_pkg_version(
            pkg_name,
            upstream=DigitDotVersion('1.2.3'),
            update_cache=False
        )
    assert (
        "should be instance of {}"
        .format(src_version_cls(pkg_name)) in str(excinfo.value)
    )


@pytest.mark.parametrize(
    'pkg_name,upstream,output,expected',
    [
        # some top level package
        ('any_package', None, '', None),
        ('any_package', None, 'Version: 1.2.3\nVersion: 1.2.4', '1.2.4'),
        ('any_package', None, 'Version: 1.2.4\nVersion: 1.2.3', '1.2.4'),
        # self package (APP_NAME)
        (APP_NAME, None, 'Version: 1.2.3\nVersion: 1.2.4', '1.2.4'),
        (APP_NAME, None, 'Version: 1.2.4\nVersion: 1.2.3', '1.2.4'),
        (APP_NAME, None, 'Version: 1.2.4~dev1\nVersion: 1.2.4~rc1', '1.2.4rc1'),
        (APP_NAME, None, 'Version: 1.2.4~rc1\nVersion: 1.2.4~dev1', '1.2.4rc1'),
        (APP_NAME, None, 'Version: 1.2.4~dev1\nVersion: 1.2.4', '1.2.4'),
        (APP_NAME, None, 'Version: 1.2.4~rc2\nVersion: 1.2.4', '1.2.4'),
        (APP_NAME, '1.2.5', 'Version: 1.2.4', None),
        (APP_NAME, '1.2.5', 'Version: 1.2.5~rc1', None),
        (APP_NAME, '1.2.5', 'Version: 1.2.5~dev1', None),
        # invalid versions from output
        ('any_package', None, 'Version: 1.2.3.4.5', None),
        (APP_NAME, None, 'Version: 1.2.3.4.5', None),
        # combined cases
        ('any_package', None, 'Version: 1.2.3\nVersion: 1.2.4\nVersion: 1.2.3.4.5', '1.2.4'),
        ('any_package', '1.2.5', 'Version: 1.2.3\nVersion: 1.2.4\nVersion: 1.2.3.4.5', None),
        (APP_NAME, None, 'Version: 1.2.3\nVersion: 1.2.4\nVersion: 1.2.5~rc1\nVersion: 1.2.5~dev1\nVersion: 1.2.3.4.5', '1.2.5rc1'),
        (APP_NAME, '1.2.5', 'Version: 1.2.3\nVersion: 1.2.4\nVersion: 1.2.5~rc1\nVersion: 1.2.5~dev1\nVersion: 1.2.3.4.5', None),
    ],
    ids=lambda s: s.replace('\n', '_').replace(' ', '_') if s else None
)
def test_get_latest_pkg_version(
        monkeypatch, pkg_name, upstream, output, expected):

    def _f(command, *args, **kwargs):
        if not output:
            raise ShellError(1, command)
        else:
            return output

    if upstream is not None:
        upstream = src_version_cls(pkg_name)(upstream)

    expected = None if expected is None else src_version_cls(pkg_name)(expected)

    monkeypatch.setattr(NodeControlUtil, 'run_shell_script_extended', _f)
    res = NodeControlUtil.get_latest_pkg_version(
        pkg_name, upstream, update_cache=False)
    assert expected == res if expected is None else res.upstream


def test_get_latest_pkg_version_for_unknown_package():
    assert NodeControlUtil.get_latest_pkg_version(
        'some-unknown-package-name', update_cache=False) is None


def test_curr_pkg_info_no_data(monkeypatch):
    monkeypatch.setattr(NodeControlUtil, 'run_shell_command', lambda *_: '')
    assert (None, []) == NodeControlUtil.curr_pkg_info('any_package')


def test_curr_pkg_info(monkeypatch):
    output = 'Version: 1.2.3\nDepends: aaa (= 1.2.4), bbb (>= 1.2.5), ccc, aaa'
    expected_deps = ['aaa=1.2.4', 'bbb=1.2.5', 'ccc']
    monkeypatch.setattr(NodeControlUtil, 'run_shell_command', lambda *_: output)

    for pkg_name in [APP_NAME, 'any_package']:
        upstream_cls = src_version_cls(pkg_name)
        expected_version = DebianVersion(
            '1.2.3', upstream_cls=upstream_cls)

        pkg_info = NodeControlUtil.curr_pkg_info(pkg_name)

        assert expected_version == pkg_info[0]
        assert isinstance(expected_version, type(pkg_info[0]))
        assert isinstance(expected_version.upstream, type(pkg_info[0].upstream))
        assert expected_deps == pkg_info[1]
